from moai.monads.execution.cascade import _create_accessor

import albumentations
import torch
import hydra.utils as hyu
import omegaconf.omegaconf
import typing
import logging
import numpy as np

log = logging.getLogger(__name__)

__all__ = ["Custom"]

class Custom(torch.utils.data.Dataset):
    def __init__(self,
        dataset:        torch.utils.data.Dataset,        
        inputs:         typing.Sequence[str],        
        outputs:        typing.Sequence[str],
        keys:           typing.Sequence[str],
        extra:          typing.List[typing.Mapping[str, str]]=[],
        augmentations:  omegaconf.DictConfig={},
    ):
        super().__init__()
        self.inner, self.inputs, self.outputs, self.keys, self.extra =\
            dataset, inputs, outputs, keys, extra
        # self.inputs = [_create_accessor(k) for k in self.inputs]
        self.inputs = [_create_accessor(k) if isinstance(k, str) else [_create_accessor(kk) for kk in k] for k in self.inputs]
        self.extra = [{k: _create_accessor(v) for k, v in (e or {}).items()} for e in self.extra]
        self.augmentations = []
        for k, v in augmentations.items():
            aug = hyu.instantiate(v)
            log.info(f"Using {k} augmentation /w a {aug.p * 100.0}% probability.")
            self.augmentations.append(aug)
        self.composition = albumentations.Compose(self.augmentations)

    def __len__(self) -> int:
        return len(self.inner)

    def __getitem__(self, index: int) -> typing.Dict[str, torch.Tensor]:
        item = self.inner[index]
        for i, o, k, e in zip(self.inputs, self.outputs, self.keys, self.extra):
            if isinstance(k, str): 
                data = {k: i(item).numpy() if isinstance(i(item),torch.Tensor) else i(item)}# numpy conversion overhead is insignificant
                for ek, ev in e.items():
                    data[ek] = ev(item).numpy() if isinstance(ev(item),torch.Tensor) \
                               else ev(item)
                augmented = self.composition(**data)
                if augmented[k] is not None:
                    item[o] = torch.from_numpy(augmented[k]) if isinstance(augmented[k],np.ndarray) else augmented[k] 
                    #item[o] = augmented[k] if isinstance(augmented[k],dict) else torch.from_numpy(augmented[k]) 
            else: # list
                data = {kk: i[j](item).numpy() if isinstance(i[j](item),torch.Tensor) else i[j](item) for j, kk in enumerate(k)}
                for ek, ev in e.items():
                    data[ek] = ev(item).numpy() if isinstance(ev(item),torch.Tensor) \
                             else ev(item)
                augmented = self.composition(**data)
                for kk, oo in zip(k, o):
                    if augmented[kk] is not None:
                        #item[oo] = torch.from_numpy(augmented[kk])
                        item[oo] = torch.from_numpy(augmented[kk])\
                            if isinstance(augmented[kk], np.ndarray) else augmented[kk]
                        # item[oo] = augmented[kk] if isinstance(augmented[kk],dict) else \
                        #     torch.from_numpy(augmented[kk])
        return item


from moai.monads.geometry.project import Projection
from moai.monads.geometry.transform_points import Transformation
from moai.monads.geometry.transform_ops import (
    Relative,
    Inverse,
    Transpose
)
from moai.monads.geometry.camera import WeakPerspective as WeakPerspectiveCamera
from moai.monads.geometry.rotate import Rotate
from moai.monads.geometry.opengl import Camera as CameraOpenGL
from moai.monads.geometry.stereo_depth import DepthFromStereo
from moai.monads.geometry.deproject import Deprojection
from moai.monads.geometry.normal_estimation2d import NormalEstimation2d

__all__ = [
    "Projection",
    "Transformation",
    "Relative",
    "Inverse",
    "Transpose",
    "WeakPerspectiveCamera",
    "Rotate",
    'CameraOpenGL',
    "DepthFromStereo",
    "Deprojection",
    "NormalEstimation2d",
]


import torch
import hydra.utils as hyu
import omegaconf.omegaconf
import typing
import toolz

__all__ = ["Concatenated"]

class Concatenated(torch.utils.data.Dataset):
    def __init__(self,
        datasets:       omegaconf.DictConfig,
        augmentation:   omegaconf.DictConfig=None,
        extracted_keys: typing.List[str]=[],
    ):
        super().__init__()
        self.datasets = []
        if augmentation is not None:
            for dataset in datasets.values():
                self.datasets.append(hyu.instantiate(
                    augmentation,
                    hyu.instantiate(dataset)
                ))
        else:
            from moai.data.augmentation import NoOp
            for dataset in datasets.values():
                self.datasets.append(NoOp(hyu.instantiate(dataset)))
        self.dataset = torch.utils.data.ConcatDataset(self.datasets)
        self.keys = [k.split('.') for k in extracted_keys]

    def __len__(self) -> int:
        return len(self.dataset)

    def __getitem__(self, index: int) -> typing.Dict[str, torch.Tensor]:
        item = self.dataset[index]
        if self.keys:            
            out = {}
            for k in self.keys:
                out = toolz.assoc_in(out, k, toolz.get_in(k, item))
            return out
        else:
            return item


from moai.monads.generation.grid.equirectangular.grid import Equirectangular as EquirectangularGrid
from moai.monads.generation.grid.equirectangular.distortion import Distortion as EquirectangularDistortion

__all__ = [
    'EquirectangularGrid',
    'EquirectangularDistortion',
]


from moai.monads.tensor.torch import (
    Concat,
    Stack,
    Split,
    SelectTensor,
    Detach,
    Flatten,
    ReshapeAs,
    Slice,
)

__all__ = [
    'Concat',
    'Stack',
    'Split',
    'SelectTensor',
    'Detach',
    'Flatten',
    'ReshapeAs',
    'Slice',
]


from moai.data.datasets.common import load_npz_file

import torch
import glob
import os
import typing
import logging

__all__ = ["Npz"]

log = logging.getLogger(__name__)

class Npz(torch.utils.data.Dataset):            #TODO add the loading different npz and combining them in the same dict case
    def __init__(self,
        root:           str='',
    ):
        self.files = glob.glob(os.path.join(root, '*.npz'))
        log.info(f"Loaded {len(self)} .npz files.")

    def __len__(self) -> int:
        return len(self.files)

    def __getitem__(self, index: int) -> typing.Dict[str, torch.Tensor]:
        return load_npz_file(self.files[index])


from moai.engine.callbacks.model import ModelCallbacks
from moai.engine.callbacks.engine import EngineCallbacks
from moai.engine.callbacks.interpolate_latent import LatentInterp
from moai.engine.callbacks.loss_schedule import LossSchedule

__all__ = [
    'ModelCallbacks',
    'EngineCallbacks',
    'LatentInterp',
    'LossSchedule',
]


import torch

__all__ = ['Deprojection']

'''
    Cartesian coordinates extraction from Spherical coordinates
        z is forward axis
        y is the up axis
        x is the right axis
        r is the radius (i.e. spherical depth)
        phi is the longitude/azimuthial rotation angle (defined on the x-z plane)
        theta is the latitude/elevation rotation angle (defined on the y-z plane)
'''
class Deprojection(torch.nn.Module):
    def __init__(self,

    ):
        super(Deprojection, self).__init__()

    def _phi(self, sgrid): # longitude or azimuth
        return sgrid[:, 0, :, :].unsqueeze(1)

    def _theta(self, sgrid): # latitude or elevation
        return sgrid[:, 1, :, :].unsqueeze(1)

    def _coord_x(self, sgrid, depth):
        return ( # r * sin(phi) * sin(theta) -> r * cos(phi) * -cos(theta) in our offsets
            depth # this is due to the offsets as explained below
            * torch.cos(self._phi(sgrid)) # long = x - 3 * pi / 2
            * -1.0 * torch.cos(self._theta(sgrid)) # lat = y - pi / 2
        )

    def _coord_y(self, sgrid, depth):
        return ( # r * cos(theta) -> r * sin(theta) in our offsets
            depth # this is due to the offsets as explained below
            * torch.sin(self._theta(sgrid)) # lat = y - pi / 2
        )

    def _coord_z(self, sgrid, depth):
        return ( # r * cos(phi) * sin(theta) -> r * -sin(phi) * -cos(theta) in our offsets
            depth # this is due to the offsets as explained above
            * torch.sin(self._phi(sgrid)) # * -1
            * torch.cos(self._theta(sgrid)) # * -1
        ) # the -1s cancel out

    def _coords_3d(self, sgrid, depth):
        return torch.cat(
            (
                self._coord_x(sgrid, depth),
                self._coord_y(sgrid, depth),
                self._coord_z(sgrid, depth)
            ), dim=1
        )

    def forward(self, 
        depth: torch.Tensor, 
        sgrid: torch.Tensor
    ) -> torch.Tensor:
        return self._coords_3d(sgrid, depth)



from moai.monads.human.body.prior.human_body_prior import (
    VPoser_v1,
    _matrot2aa,
)

import torch
import typing

#NOTE: code from https://github.com/nghorbani/human_body_prior

__all__ = ["VPoser1"]

class VPoser1(VPoser_v1):
    def __init__(self,
        flatten_pose:       bool=True,
    ):
        super().__init__(
            num_neurons=512,
            latentD=32,
            data_shape=[1, 21, 3],
            use_cont_repr=True,
        )
        self.eval()
        self.flatten_pose = flatten_pose
    
    def forward(self,
        encode:         torch.Tensor=None,
        decode:         torch.Tensor=None,
        autoencode:     torch.Tensor=None,
    ) -> typing.Mapping[str, torch.Tensor]:
        out = { }        
        if autoencode is not None:
            # decoded = super(VPoser2, self).forward(autoencode)
            q_z = self.encode(autoencode)
            # q_z_sample = q_z.rsample()
            q_z_sample = q_z.mean
            dec = self.decode(q_z_sample)
            bs = autoencode.shape[0]
            decoded = {
                'pose_body': _matrot2aa(dec.view(-1, 3, 3)).view(bs, -1, 3),
                'pose_body_matrot': dec.view(bs, -1, 9),
            }
            decoded.update({
                'poZ_body_mean': q_z.mean,
                'poZ_body_std': q_z.scale,
                'q_z': q_z,
            })
            out['pose'] = decoded['pose_body']
            out['embedding'] = decoded['poZ_body_mean'] # decoded['q_z']
            if self.flatten_pose:
                out['pose'] = out['pose'].reshape(autoencode.shape[0], -1)
            return out
        if encode is not None:
            out['embedding'] = self.encode(encode)
        if decode is not None:
            dec = self.decode(decode)
            bs = decode.shape[0]            
            out['pose'] = _matrot2aa(dec.view(-1, 3, 3)).view(bs, -1, 3)
            if self.flatten_pose:
                out['pose'] = out['pose'].reshape(decode.shape[0], -1)
        return out


import torch
import typing
import functools
import numpy as np
import logging

#NOTE: code from https://github.com/vchoutas/smplify-x

__all__ = [
    "Split",
    "JointMap",
    "JointConfidence",
    "MergeToes",
]

log = logging.getLogger(__name__)

__JOINT_FORMATS__ = {
    'none':                 [118,   0,          0],
    'coco25':               [25,    21 * 2,     51],
    'coco25_face':          [25,    21 * 2,     51 + 17],
    'coco25_star':          [25,    0,          0],
    'coco25_star+':         [25,    21 * 2,     0],
}

class Split(torch.nn.Module):
    def __init__(self,
        format:             str='coco25',
    ):
        super(Split, self).__init__()
        self.split_sections = __JOINT_FORMATS__[format]

    def forward(self,
        joints:          torch.Tensor=None,
    ) -> typing.Mapping[str, torch.Tensor]:
        s = sum(self.split_sections)
        body, hands, face = torch.split(joints[..., :s, :], self.split_sections, dim=-2)
        return {
            'body':     body,
            'hands':    hands,
            'face':     face,
        }

def _body_to_openpose(
    model_type:         str='smplx',
    openpose_format:    str='coco25',
    use_hands:          bool=True,
    use_face:           bool=True,
    use_face_contour:   bool=False,    
) -> np.ndarray:
    ''' Returns the indices of the permutation that maps OpenPose to SMPL

        Parameters
        ----------
        model_type: str, optional
            The type of SMPL-like model that is used. The default mapping
            returned is for the SMPLX model
        use_hands: bool, optional
            Flag for adding to the returned permutation the mapping for the
            hand keypoints. Defaults to True
        use_face: bool, optional
            Flag for adding to the returned permutation the mapping for the
            face keypoints. Defaults to True
        use_face_contour: bool, optional
            Flag for appending the facial contour keypoints. Defaults to False
        openpose_format: bool, optional
            The output format of OpenPose. For now only COCO-25 and COCO-19 is
            supported. Defaults to 'coco25'

    '''
    if openpose_format.lower() == 'coco25':
        if model_type == 'smpl':
            return np.array([
                24, 12, 17, 19, 21,
                16, 18, 20, 0, 2,
                5, 8, 1, 4, 7,
                25, 26, 27, 28, 29,
                30, 31, 32, 33, 34
            ], dtype=np.int32)        
        elif model_type == 'smplh':
            body_mapping = np.array([
                52, 12, 17, 19, 21,
                16, 18, 20, 0, 2,
                5, 8, 1, 4, 7,
                53, 54, 55, 56, 57,
                58, 59, 60, 61, 62
            ], dtype=np.int32)
            mapping = [body_mapping]
            if use_hands:
                lhand_mapping = np.array([
                    20, 34, 35, 36, 63,
                    22, 23, 24, 64, 25,
                    26, 27, 65, 31, 32,
                    33, 66, 28, 29, 30,
                    67
                ], dtype=np.int32)
                rhand_mapping = np.array([
                    21, 49, 50, 51, 68,
                    37, 38, 39, 69, 40,
                    41, 42, 70, 46, 47, 
                    48, 71, 43, 44, 45,
                    72
                ], dtype=np.int32)
                mapping += [lhand_mapping, rhand_mapping]
            return np.concatenate(mapping)
        # SMPLX
        elif model_type == 'smplx':
            body_mapping = np.array([
                55, 12, 17, 19, 21,
                16, 18, 20, 0, 2,
                5, 8, 1, 4, 7, 
                56, 57, 58, 59, 60,
                61, 62, 63, 64, 65
            ], dtype=np.int32)
            mapping = [body_mapping]
            if use_hands:
                lhand_mapping = np.array([
                    20, 37, 38, 39, 66,
                    25, 26, 27, 67, 28,
                    29, 30, 68, 34, 35,
                    36, 69, 31, 32, 33, 
                    70
                ], dtype=np.int32)
                rhand_mapping = np.array([
                    21, 52, 53, 54, 71,
                    40, 41, 42, 72, 43,
                    44, 45, 73, 49, 50,
                    51, 74, 46, 47, 48,
                    75
                ], dtype=np.int32)
                mapping += [lhand_mapping, rhand_mapping]
            if use_face:
                #  end_idx = 127 + 17 * use_face_contour
                face_mapping = np.arange(76, 127 + 17 * use_face_contour, dtype=np.int32)
                mapping += [face_mapping]
            return np.concatenate(mapping)
        elif model_type == 'star':
            body_mapping = np.array([
                15, 12, 17, 19, 21, 
                16, 18, 20, 0, 2, 
                5, 8, 1, 4, 7, 
                24, 25, 26, 27, 10,
                10, 7, 11, 11, 8,
            ], dtype=np.int32)
            mapping = [body_mapping]
            if use_hands:
                lhand_mapping = np.array([
                    0, 0, 0, 0, 0,
                    0, 0, 0, 0, 22,
                    0, 0, 0, 0, 0,
                    0, 0, 0, 0, 0,
                    0
                ], dtype=np.int32)
                rhand_mapping = np.array([
                    0, 0, 0, 0, 0,
                    0, 0, 0, 0, 23,
                    0, 0, 0, 0, 0,
                    0, 0, 0, 0, 0,
                    0
                ], dtype=np.int32)
                mapping += [lhand_mapping, rhand_mapping]
            if use_face: 
                #NOTE: here we place the face landmarks to the original body mapping
                mapping[0][15] = 25 # REye
                mapping[0][16] = 26 # LEye
                mapping[0][17] = 24 # REar
                mapping[0][18] = 27 # LEar
                #NOTE: currently not used anywhere, just a placeholder for the face landmarks
                # face_mapping = np.arange(76, 127 + 17 * use_face_contour, dtype=np.int32)
                # mapping += [face_mapping]
            return np.concatenate(mapping)
        else:
            log.error(f'Unknown model type: {model_type}')
    elif openpose_format == 'coco19':
        if model_type == 'smpl':
            return np.array([
                24, 12, 17, 19, 21,
                16, 18, 20, 0, 2,
                5, 8, 1, 4, 7,
                25, 26, 27, 28
            ], dtype=np.int32)
        elif model_type == 'smplh':
            body_mapping = np.array([
                52, 12, 17, 19, 21,
                16, 18, 20, 0, 2, 
                5, 8, 1, 4, 7,
                53, 54, 55, 56
            ], dtype=np.int32)
            mapping = [body_mapping]
            if use_hands:
                lhand_mapping = np.array([
                    20, 34, 35, 36, 57,
                    22, 23, 24, 58, 25,
                    26, 27, 59, 31, 32,
                    33, 60, 28, 29, 30,
                    61
                ], dtype=np.int32)
                rhand_mapping = np.array([
                    21, 49, 50, 51, 62,
                    37, 38, 39, 63, 40,
                    41, 42, 64, 46, 47,
                    48, 65, 43, 44, 45,
                    66
                ], dtype=np.int32)
                mapping += [lhand_mapping, rhand_mapping]
            return np.concatenate(mapping)
        # SMPLX
        elif model_type == 'smplx':
            body_mapping = np.array([
                55, 12, 17, 19, 21,
                16, 18, 20, 0, 2,
                5, 8, 1, 4, 7,
                56, 57, 58, 59
            ], dtype=np.int32)
            mapping = [body_mapping]
            if use_hands:
                lhand_mapping = np.array([
                    20, 37, 38, 39, 60,
                    25, 26, 27, 61, 28,
                    29, 30, 62, 34, 35,
                    36, 63, 31, 32, 33,
                    64
                ], dtype=np.int32)
                rhand_mapping = np.array([
                    21, 52, 53, 54, 65,
                    40, 41, 42, 66, 43,
                    44, 45, 67, 49, 50,
                    51, 68, 46, 47, 48,
                    69
                ], dtype=np.int32)
                mapping += [lhand_mapping, rhand_mapping]
            if use_face:
                face_mapping = np.arange(70, 70 + 51 + 17 * use_face_contour, dtype=np.int32)
                mapping += [face_mapping]
            return np.concatenate(mapping)
        else:
            log.error(f'Unknown model type: {model_type}')
    else:
        log.error(f'Unknown joint format: {openpose_format}')

class JointMap(torch.nn.Module):
    def __init__(self, 
        model:              str='smplx',
        format:             str='coco25',
        with_hands:         bool=False,
        with_face:          bool=False,
        with_face_contour:  bool=False,
    ):
        super(JointMap, self).__init__()
        self.register_buffer('indices', torch.from_numpy(
            _body_to_openpose(model_type=model, 
                openpose_format=format, 
                use_hands=with_hands,
                use_face=with_face,
                use_face_contour=with_face_contour,
            )).long()
        )
        self.index = functools.partial(torch.index_select, dim=1)

    def forward(self, 
        joints: torch.Tensor,
        vertices: torch.Tensor=None, #NOTE: for compatibility with current SMPLX implementation
    ) -> torch.Tensor:
        return self.index(joints, index=self.indices)

class JointConfidence(torch.nn.Module):
    def __init__(self,
        joints_to_ignore:       typing.Sequence[int],
        confidence_threshold:   float=0.0,
    ):
        super(JointConfidence, self).__init__()
        self.register_buffer('ignore', torch.tensor(joints_to_ignore).long())
        self.threshold = confidence_threshold

    def forward(self, 
        confidence: torch.Tensor
    ) -> torch.Tensor:
        ret = confidence.clone()
        if len(ret.shape) == 3:
            ret[:, self.ignore, ...] = 0.0
        else:
           ret[self.ignore, ...] = 0.0 
        ret[ret < self.threshold] = 0.0
        return ret

class MergeToes(torch.nn.Module): #TODO: aug layer
    def __init__(self):
        super(MergeToes, self).__init__()

    def forward(self,
        keypoints:  torch.Tensor,
        confidence: torch.Tensor,
    ) -> typing.Dict[str, torch.Tensor]:
        kpts = keypoints.detach().clone()
        conf = confidence.detach().clone()
        # right
        right_toes_j = kpts[:, 22:24, :]
        right_toes_w = conf[:, 22:24, :]
        right_toe_w = right_toes_w.sum(dim=1, keepdim=True)
        right_toe = (right_toes_j * right_toes_w).sum(dim=1, keepdim=True) / (right_toe_w + 1e-8)
        kpts[:, 22, :] = right_toe
        kpts[:, 23, :] = right_toe
        right_toe_w = right_toe_w * 0.5
        conf[:, 22, :] = right_toe_w
        conf[:, 23, :] = right_toe_w
        # left
        left_toes_j = kpts[:, 19:21, :]
        left_toes_w = conf[:, 19:21, :]
        left_toe_w = left_toes_w.sum(dim=1, keepdim=True)
        left_toe = (left_toes_j * left_toes_w).sum(dim=1, keepdim=True) / (left_toe_w + 1e-8)
        kpts[:, 19, :] = left_toe
        kpts[:, 20, :] = left_toe
        left_toe_w = left_toe_w * 0.5
        conf[:, 19, :] = left_toe_w
        conf[:, 20, :] = left_toe_w
        return {
            'positions' : kpts,
            'confidence': conf,
        }



from pytorch_lightning.callbacks import Callback
import numpy as np
import logging
import torch

log = logging.getLogger(__name__)

__all__ = ["LatentInterp"]

class LatentInterp(Callback):
    def __init__(self,
        num_points: int,
        interpolate_epoch_interval: int,
        range_start: int=-5,
        range_end: int=5,
        steps: int=11,
        num_samples: int=1,
        normalize: bool=True,
        full_vector: bool=False,
    ):
        self.interpolate_epoch_interval = interpolate_epoch_interval
        self.range_start = range_start
        self.range_end = range_end
        self.num_samples = num_samples
        self.num_points = num_points
        self.normalize = normalize
        self.steps = steps
        self.full_vector = full_vector

    def on_validation_epoch_start(self, trainer, pl_module):
        points = self.interpolate_latent(pl_module, pl_module.latent_dim)
        pl_module.visualization.visualizers[2](points.reshape(self.steps, self.num_points, 3)) if self.full_vector\
                                        else pl_module.visualization.visualizers[2](points.reshape(self.steps*self.steps, self.num_points, 3))
    
    def interpolate_latent(self, pl_module, latent_dim):
        with torch.no_grad():
            pl_module.eval()
            points_list = []
            if self.full_vector:
                z = torch.randn(self.num_samples, latent_dim, device=pl_module.device)
                for full_z in np.linspace(self.range_start, self.range_end, self.steps):
                    z[:] += torch.tensor(full_z)
                    z_reshaped = pl_module.reparametrizer.linear_to_dec(z)
                    #TODO solve the reshape problem
                    points = pl_module.decoder(z_reshaped)
                    points_list.append(points)
            else:
                z = torch.randn(self.num_samples, latent_dim, device=pl_module.device)
                for z1 in np.linspace(self.range_start, self.range_end, self.steps):
                    for z2 in np.linspace(self.range_start, self.range_end, self.steps):

                        # set the fist 2 dims to the value
                        z[:, 0] = torch.tensor(z1)
                        z[:, 1] = torch.tensor(z2)

                        # generate
                        z_reshaped = pl_module.reparametrizer.linear_to_dec(z)
                        #TODO solve the reshape problem
                        points = pl_module.decoder(z_reshaped)
                        points_list.append(points)
        return torch.stack(points_list, 1).squeeze(0)


# Copyright 2019 The Google Research Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from moai.supervision.losses.regression.robust.adaptive.distribution import Distribution
from moai.supervision.losses.regression.robust.adaptive import util
from moai.utils.arguments.common import (
  assert_non_negative,
  assert_numeric,
)

import numpy as np
import torch
import typing
import logging

log = logging.getLogger(__name__)

__all__ = [
    'Barron',
    'Student',
]

class Barron(Distribution):
    def __init__(self,
        count:          int,
        alpha_range:    typing.Union[float, typing.Tuple[float, float]]=[0.001, 1.999],
        scale_low:      float=1e-5,
        scale_init:     float=1.0,
        alpha_init:     typing.Optional[float]=None,
    ):
      super().__init__()
      alpha_range = sorted(alpha_range)
      assert_numeric(log, 'alpha low', alpha_range[0], 0.0, 2.0)
      assert_numeric(log, 'alpha high', alpha_range[1], 0.0, 2.0)
      assert_non_negative(log, 'scale low', scale_low)
      if alpha_range[0] == alpha_range[1]: # constant init alpha
          log.info(f"Barron loss alpha set to fixed: {alpha_range[0]}.")
          self.fixed_alpha = torch.tensor(
              alpha_range[0])[np.newaxis, np.newaxis].repeat(1, count)
          self.alpha = lambda: self.fixed_alpha
      else: # learnable alpha
          alpha = util.inverse_affine_sigmoid(
              np.mean(alpha_range) if alpha_init is None else alpha_init,
              low=alpha_range[0], high=alpha_range[1]
          )
          self.register_parameter('alpha', torch.nn.Parameter(
                  alpha.clone().detach().float()[np.newaxis, np.newaxis].repeat(1, count),
                  requires_grad=True
              )
          )
          self.get_alpha = lambda: util.affine_sigmoid(
              self.alpha, low=alpha_range[0], high=alpha_range[1]
          )
      if scale_low == scale_init: # fix `scale` to be a constant.
        log.info(f"Barron loss scale set to fixed: {scale_init}.")
        self.register_buffer('fixed_scale', 
          torch.tensor(scale_init)[np.newaxis, np.newaxis].float().repeat(1, count)
        )
        self.get_scale = lambda: self.fixed_scale
      else: # learnable scale
        self.register_parameter('scale', torch.nn.Parameter(
                torch.zeros((1, count)).float(),
                requires_grad=True
            )
        )
        self.get_scale = lambda: util.affine_softplus(
            self.scale, low=scale_low, ref=scale_init
        )
      #TODO: use parameterizations instead of lambdas

    def forward(self, 
        gt:   torch.Tensor,
        pred: torch.Tensor
    ) -> torch.Tensor:
        return super().forward(
            pred.flatten(1) - gt.flatten(1), 
            self.get_alpha(), 
            self.get_scale()
        )

class Student(torch.nn.Module):
    def __init__(self,
        count:      int,
        scale_low:  float=1e-5,
        scale_init: float=1.0,
    ):
      super(Student, self).__init__()
      assert_non_negative(log, "scale low", scale_low)
      self.register_parameter('log_df', torch.nn.Parameter(
          torch.zeros((1, count)).float(),
          requires_grad=True
      ))
      self.df = lambda: torch.exp(self.log_df)
      if scale_low == scale_init: # fix `scale` to be a constant.
          self.register_buffer('scale', 
            torch.tensor(scale_init).float()[np.newaxis, np.newaxis].repeat(1, count)
          )
          self.get_scale = lambda: self.scale
      else: # learnable scale
          self.register_parameter('scale', torch.nn.Parameter(
                torch.zeros((1, count)).float(), requires_grad=True
          ))
          self.get_scale = lambda: util.affine_softplus(
              self.scale, low=scale_low, ref=scale_init
          )
      #NOTE: explore parameterizations
      
    def lossfun(self, 
        gt:   torch.Tensor,
        pred: torch.Tensor,
    ) -> torch.Tensor:
        return util.students_t_nll(
          pred.flatten(1) - gt.flatten(1),
          self.df(), 
          self.scale()
        )



import moai.nn.convolution as mic
import moai.nn.activation as mia
import moai.nn.linear as milin
import moai.nn.sampling.spatial.downsample as mids

from moai.networks.lightning.factory import HRNet, StackedHourglass

import torch
import omegaconf.omegaconf
import functools
import toolz
import typing

__all__ = ['SqueezeExcite']

#NOTE: from https://github.com/xvjiarui/GCNet/blob/029db5407dc27147eb1d41f62b09dfed8ec88837/mmdet/ops/gcb/context_block.py#L64
class Attention2d(torch.nn.Module):
    def __init__(self,
        features:       int,
    ):
        super(Attention2d, self).__init__()
        self.conv = torch.nn.Conv2d(features, 1, kernel_size=1)
        self.softmax = torch.nn.Softmax(dim=2)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        b, c, h, w = x.shape        
        y = x.view(b, c, h * w).unsqueeze(1)
        mask = self.conv(x)        
        mask = torch.nn.functional.softmax(
            mask.view(b, 1, h * w), dim=-1
        ).unsqueeze(-1)
        context = torch.matmul(y, mask)
        return context.view(b, c, 1, 1)

class GlobalAveragePool(torch.nn.Module):
    def __init__(self,   
        dims:       int,
        features:   int,        
    ):
        super(GlobalAveragePool, self).__init__()
        self.pool_func = functools.partial(
            getattr(torch.nn.functional, f"adaptive_avg_pool{dims}d"),
            output_size=1
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.pool_func(x)

__SQUEEZE__ = {
    'average1d':    functools.partial(GlobalAveragePool, dims=1),
    'average2d':    functools.partial(GlobalAveragePool, dims=2),
    'average3d':    functools.partial(GlobalAveragePool, dims=3),
    'attention2d':  Attention2d,
    'none':         torch.nn.Identity,
}

def make_channel_conv(
    type:               str,
    in_features:        int,
    out_features:       int,
    operation_params:   dict,
    activation_type:    str,
    activation_params:  dict,
):
    return mic.make_conv_block(
        block_type=type,
        convolution_type=type, 
        in_features=in_features, 
        out_features=out_features,
        activation_type=activation_type,
        convolution_params=toolz.merge(
            operation_params, 
            {'kernel_size': 1, 'padding': 0, 'stride': 1}
        ),
        activation_params=toolz.merge({'inplace': True}, activation_params)
    )

def make_channel_linear(
    type:               str,
    in_features:        int,
    out_features:       int,
    operation_params:   dict,
    activation_type:    str,
    activation_params:  dict,
):
    return milin.make_linear_block(
        block_type=type,
        linear_type=type, 
        in_features=in_features, 
        out_features=out_features,
        activation_type=activation_type,
        linear_params=operation_params,
        activation_params=toolz.merge({'inplace': True}, activation_params),
    )

__OPERATIONS__ = {
    'conv2d':   functools.partial(make_channel_conv, type='conv2d'),
    'linear':   functools.partial(make_channel_linear, type='linear'),
}

def make_channel(
    operation_type:     str,
    activation_type:    str,
    in_features:        int,    
    operation_params:   dict=None,    
    activation_params:  dict=None,
    ratio:              typing.Union[float, int]=0.5,
):
    inter_features = int(in_features * ratio)\
        if isinstance(ratio, float) else in_features // ratio
    return torch.nn.Sequential(
            __OPERATIONS__[operation_type](
                in_features=in_features,
                out_features=inter_features, 
                operation_params=operation_params or { },
                activation_type=activation_type or 'none',
                activation_params=activation_params or { }
            ),
            __OPERATIONS__[operation_type](
                in_features=inter_features,
                out_features=in_features,
                operation_params=operation_params or { },
                activation_type='none',
                activation_params={ }
            ),
            torch.nn.Unflatten(1, (in_features, 1, 1)) \
                if 'linear' in operation_type else torch.nn.Identity()
        )

def make_spatial(
    operation_type:     str,
    activation_type:    str,
    in_features:        int,    
    operation_params:   dict=None,
    activation_params:  dict=None,
    ratio:              float=1.0,
):
    return make_channel_conv(
        type=operation_type,
        in_features=in_features,
        out_features=1,
        operation_params=operation_params or { },
        activation_type=activation_type,
        activation_params=activation_params or { },        
    )

__EXCITE__ = {
    'channel': make_channel,
    'spatial': make_spatial,
    #TODO: 'channel_spatial': #NOTE: with max of channel and spatial, see https://github.com/ai-med/squeeze_and_excitation/blob/acdd26e7e3956b8e3d3b32663a784ebd64c844dd/squeeze_and_excitation/squeeze_and_excitation_3D.py#L119
}

#NOTE: it is an activation
#TODO: check with bias as well: https://github.com/JYPark09/SENet-PyTorch/blob/6f1eae93256e5181baea8d5102473c6cba6500fa/network.py#L6
class SqueezeExcite(torch.nn.Module):
    def __init__(self,
        features:       int,
        squeeze:        omegaconf.DictConfig, # type: 'averageXd', # one of ['averageXd', 'attentionXd']
        excite:         omegaconf.DictConfig, # delta: activation_{type|params}, # ratio \in [0, 1], # operation_{type|params}: one of ['convXd', 'linear']
        mode:           str='mul',
    ):
        super(SqueezeExcite, self).__init__()
        self.squeeze = __SQUEEZE__[squeeze.type](features=features)
        self.excite = __EXCITE__[excite.type](
            operation_type=excite.operation.type,
            activation_type=toolz.get_in(['activation', 'type'], excite, 'none'),
            in_features=features,
            operation_params=excite.operation.params or { },
            activation_params=toolz.get_in(['activation', 'params'], excite, { }),
            ratio=excite.ratio,
        )
        self.mode = getattr(torch, mode)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        squeezed = self.squeeze(x)
        excited = self.excite(squeezed)
        return self.mode(torch.sigmoid(excited), x)

if __name__ == '__main__':
    FEATURES = 32
    t = torch.rand(5, FEATURES, 24, 32)
    se = SqueezeExcite(
        features=FEATURES,
        # ''' TRADITIONAL Squeeze-n-Excite '''
        squeeze=omegaconf.DictConfig({
            'type': 'average2d',
            # 'type': 'attention2d',
        }),        
        excite=omegaconf.DictConfig({
            'type': 'channel',
            'operation': {
                # 'type': 'linear',
                'type': 'conv2d',
            },
            'activation': {
                'type': 'bn2d_relu',
                # 'type': 'relu_bn2d',
                # 'type': 'relu',
            },
            'ratio': 16, # 0.25,
        }),
        # ''' Spatial Squeeze-n-Excite '''
        # squeeze=omegaconf.DictConfig({
        #     'type': 'none',
        # }),        
        # excite=omegaconf.DictConfig({
        #     'type': 'spatial',
        #     'operation': {
        #         'type': 'conv2d',
        #         'params': {
        #             'bias': False,
        #         }
        #     },
        # }),
        # mode='mul',        
    )
    res = se(t)
    print(res.shape)


from moai.monads.human.pose.mediapipe import JointMap

import toolz
import torch
import smplx #TODO: try/except and error msg
import functools
import typing

#NOTE: code from https://github.com/vchoutas/smplify-x

__all__ = ["MANO"]

__JOINT__MAPPERS__ = {
    'none':             None,
    'mediapipe_hand':  functools.partial(JointMap, model='mano', format='mediapipe_hand'),
}

#NOTE: joints: https://github.com/otaheri/MANO/blob/master/mano/joints_info.py
#NOTE: see https://github.com/vchoutas/smplx/issues/48
class MANO(smplx.MANO):
    def __init__(self,
        model_path:             str,
        joints_format:          str='none',                
        is_right_hand:          bool=True,
        num_betas:              int=10,
        pca_components:         int=6,
        use_global_orientation: bool=True, # root rotation
        use_translation:        bool=True, # body translation
        use_pose:               bool=True, # joint rotations, False when using VPoser
        use_betas:              bool=True, # shape        
        use_pca:                bool=True,
        flat_hand_mean:         bool=False,
        with_tips:              bool=False,     
    ):
        self.with_tips = with_tips
        super(MANO, self).__init__(
            model_path=model_path,
            joint_mapper=None,
            create_global_orient=use_global_orientation,
            create_hand_pose=use_pose,
            create_betas=use_betas,
            create_transl=use_translation,
            use_pca=use_pca,
            dtype=torch.float32,
            batch_size=1,
            num_pca_comps=pca_components,
            num_betas=num_betas,
            flat_hand_mean=flat_hand_mean,
            is_rhand=is_right_hand
        )

    def forward(self,
        betas:          torch.Tensor=None,
        pose:           torch.Tensor=None,
        rotation:       torch.Tensor=None,
        translation:    torch.Tensor=None,
    ) -> typing.Mapping[str, torch.Tensor]:
        if pose is not None and len(pose.shape) > 2:
            pose = torch.flatten(pose, start_dim=1)
        hand_output = super(MANO, self).forward(
            betas=betas,                    # betas -> [1, 10] # v_shaped -> [1, 778, 3]
            hand_pose=pose,                 # body_pose -> [1, 45] or [1, 12] # joints -> [1, 15, 3]
            global_orient=rotation,         # global_orient -> [1, 3]
            transl=translation,             # transl -> [1, 3]
            return_full_pose=True,          # full_pose -> [1, 48] => 15 joints * 3 + 3 * global rotation
            return_verts=True,              # vertices -> [1, 778, 3]
        )
        b = betas.shape[0]
        if self.with_tips:
            joints = torch.cat(
                [
                    hand_output['joints'],
                    hand_output['vertices'][:,744:745,:], #thumb 16
                    hand_output['vertices'][:,320:321,:], #index 17
                    hand_output['vertices'][:,443:444,:], #middle 18
                    hand_output['vertices'][:,554:555,:], #ring 19
                    hand_output['vertices'][:,671:672,:], #pinky 20
                ], dim=-2)
        else:
            joints = hand_output['joints']
        return toolz.valfilter(lambda v: v is not None, {
            'vertices':     hand_output['vertices'],
            'pose':         hand_output['hand_pose'],
            'rotation':     hand_output['global_orient'],
            'translation':  hand_output['transl'],
            'betas':        hand_output['betas'],
            'shape':        hand_output['v_shaped'],
            'joints':       joints,
            'full_pose':    hand_output['full_pose'],
            'faces':        self.faces_tensor.expand(b, -1, -1), # faces [1, 1538 3]
        })

RightHandMANO = functools.partial(MANO, is_right_hand=True)
LeftHandMANO = functools.partial(MANO, is_right_hand=False)


from moai.data.datasets.human.pose.openpose import (
    OpenPoseInferred,
    OpenPoseKeypoints,
    OpenPoseMultiviewKeypoints,
)

__all__ = [
    'OpenPoseInferred',
    "OpenPoseKeypoints",
    "OpenPoseMultiviewKeypoints",
]


import numpy as np
import torch
import typing

__all__ = ['Npy']

class Npy(torch.nn.Module):
    def __init__(self,
        **kwargs: typing.Mapping[str, str],        
    ) -> None:
        super().__init__()
        for k, f in kwargs.items():
            self.register_buffer(k,
                torch.from_numpy(np.load(f)).float()
            )

    def forward(self, tensor: torch.Tensor) -> torch.Tensor:
        return { k:v for k, v in self.named_buffers() }


from torch.distributions import MultivariateNormal as MVN
import torch

import logging

log = logging.getLogger(__name__)

__all__ = [
    'ScalarMSE',
    'VectorMSE',
]

class ScalarMSE(torch.nn.Module):
    def __init__(self,
        sigma:      float,
        mode:       str='auto', # one of [auto, const]
    ):
        super(ScalarMSE, self).__init__()
        if mode == 'auto':
            self.register_parameter('sigma', torch.nn.Parameter(torch.scalar_tensor(sigma)))
        else:
            self.register_buffer('sigma', torch.nn.Parameter(torch.scalar_tensor(sigma)))

    def forward(self,        
        pred:       torch.Tensor,
        gt:         torch.Tensor=None, #NOTE: w/o gt it serves as a prior
        weights:    torch.Tensor=None, # float tensor
        mask:       torch.Tensor=None, # byte tensor
    ) -> torch.Tensor:
        p = pred.flatten(start_dim=1).reshape(-1, 1)
        g = gt.flatten(start_dim=1).reshape(-1, 1)
        noise_var = self.sigma ** 2
        logits = - 0.5 * (p - g.T).pow(2) / noise_var
        loss = torch.nn.functional.cross_entropy(logits, torch.arange(p.shape[0], device=p.device), reduction='none')
        loss = loss * (2 * noise_var).detach()        
        # if weights is not None:
        #     l2 = l2 * weights
        # if mask is not None:
        #     l2 = l2[mask]        
        return loss

class VectorMSE(torch.nn.Module):
    def __init__(self,
        sigma:      float=0.0,
        dim:        int=1,
        mode:       str='auto', # one of [auto, const]
    ):
        super(VectorMSE, self).__init__()        
        if mode == 'auto':
            self.register_parameter('sigma', torch.nn.Parameter(
                torch.scalar_tensor(sigma) if dim == 1 else torch.tensor([sigma] * dim)
            ))                
        else:
            self.register_buffer('sigma', torch.nn.Parameter(
                torch.scalar_tensor(sigma) if dim == 1 else torch.tensor([sigma] * dim)
            ))

    def forward(self,        
        pred:       torch.Tensor,
        gt:         torch.Tensor=None, #NOTE: w/o gt it serves as a prior
        weights:    torch.Tensor=None, # float tensor
        mask:       torch.Tensor=None, # byte tensor
    ) -> torch.Tensor:
        # if pred.shape[0] == 1:
        #     pred, gt = pred[0], gt[0]        
        if weights is not None:
            s = pred.shape
            ctrl = self.sigma.detach().view(*s[1:]).unsqueeze(0).expand_as(pred)
            pred = pred * weights + ctrl
            gt = gt * weights - ctrl
        if mask is not None:
            s = pred.shape
            ctrl = self.sigma.detach().view(*s[1:]).unsqueeze(0).expand_as(pred)
            pred[mask] = ctrl
            gt[mask] = -ctrl
        p = pred.flatten(start_dim=1)
        g = gt.flatten(start_dim=1)
        noise_var = self.sigma ** 2        
        I = torch.eye(p.shape[-1]).to(gt.device)
        logits = MVN(p.unsqueeze(1), noise_var * I).log_prob(g.unsqueeze(0)) # [B,1,N] / [N,N] / [1,B,N]  => [B, B]
        loss = torch.nn.functional.cross_entropy(logits, torch.arange(p.shape[0], device=p.device))
        loss = loss * (2 * noise_var).detach()             
        return loss


from moai.utils.arguments import assert_path
from moai.data.datasets.common.image2d import (
    load_color_image,
    load_mask_image,
)

import toolz
import numpy as np
import glob
import torch
import os
import typing
import logging
import trimesh

log = logging.getLogger(__name__)

__all__ = ['GeneBody']

class GeneBody(torch.utils.data.Dataset):
    def __init__(self,
        root:               str,
        split:              str='test',
        subjects:           typing.Union[typing.List[str], str]='**',
        cameras:            typing.Union[typing.List[str], str]='**', #NOTE: the default seting of GNR is to use these four source views of GeneBody: ['01', '13', '25', '37']
        # modalities:         typing.List[str]=['image'],
    ) -> None:
        super().__init__()
        assert_path(log, 'GenBody root path', root)
        is_all_subjects = isinstance(subjects, str) and ('all' == subjects or '**' == subjects)
        subjects = os.listdir(os.path.join(root, split)) if is_all_subjects else subjects
        subjects = [subjects] if isinstance(subjects, str) else subjects
        is_all_cameras = isinstance(cameras, str) and ('all' == cameras or '**' == cameras)
        cameras = os.listdir(os.path.join(root, split, 'image')) if is_all_cameras else cameras
        cameras = [cameras] if isinstance(cameras, str) else cameras
        self.metadata = { 
            'root': root, 'items': [ ], 'subjects': { },
            'cameras': np.array(cameras).astype(np.int32),
        }
        for subject in subjects:
            params = glob.glob(os.path.join(root, split, subject, 'param', '*.*'))
            cams = np.load(os.path.join(root, split, subject, 'annots.npy'), allow_pickle=True).item()['cams']
            extra = '0' if subject == 'zhuna' else ''
            for param in params:
                filename = os.path.basename(param)
                self.metadata['items'].append({
                    'smplx': param,
                    'mesh': param.replace('param', 'smpl').replace('npy', 'obj'),
                    'images': {c: os.path.join(root, split, subject, 'image', c, filename.replace('npy', 'jpg')) for c in cameras},
                    'masks': {c: os.path.join(root, split, subject, 'mask', c, 'mask' + extra + filename.replace('npy', 'png')) for c in cameras},
                    'cameras': toolz.valmap(
                        lambda d: toolz.valmap(
                            lambda v: torch.from_numpy(v).squeeze().float(), 
                            d
                        ), 
                        toolz.keyfilter(lambda k: k in cameras, cams)
                    )
                })
        log.info(f"Loaded {len(self)} items from GeneBody.")

    def __len__(self) -> int:
        return len(self.metadata['items'])

    def load_smplx(self, filename: str) -> typing.Dict[str, torch.Tensor]:
        data = np.load(filename, allow_pickle=True).item()
        return {
            'scale': torch.scalar_tensor(data['smplx_scale']).float(),             
            'params': toolz.valmap(lambda v: v.squeeze().float(),
                toolz.valfilter(lambda v: isinstance(v, torch.Tensor), data['smplx'])
            ),
            'keypoints3d': torch.from_numpy(data['smplx']['keypoints3d']).squeeze().float()
        }
        
    def __getitem__(self, index: int) -> typing.Dict[str, torch.Tensor]:
        item = self.metadata['items'][index]
        mesh = trimesh.load(item['mesh'], process=False)
        smplx = self.load_smplx(item['smplx'])
        data = {
            'smplx': toolz.merge(smplx, {
                'mesh': {
                    'vertices': torch.from_numpy(mesh.vertices).float(),
                    'scaled_vertices': smplx['scale'] * torch.from_numpy(mesh.vertices).float(),
                    'faces': torch.from_numpy(mesh.faces).int() #NOTE: or long?
                },
            }),
            'images': toolz.valmap(load_color_image, item['images']),
            'masks': toolz.valmap(load_mask_image, item['masks']),
            'cameras': item['cameras']
        }
        return data


import torch
import numpy as np

__all__ = ['DistanceAccuracy']

class DistanceAccuracy(torch.nn.Module):
    def __init__(self,
        threshold: float=0.05
    ) -> None:
        super().__init__()
        self.threshold = threshold
    
    def forward(self,
        distance: torch.Tensor #distance between gt and predicted kpts
    ) -> torch.Tensor:
        total = np.prod(list(distance.shape)[1:])
        acc = (distance.flatten(1) < self.threshold).sum(1) / total
        return acc.mean()


import torch
import typing

class Stack(torch.nn.Module):
    def __init__(self,
        dim: int=0
    ):
        super(Stack, self).__init__()
        self.dim = dim

    def forward(self, tensors: typing.List[torch.Tensor]) -> torch.Tensor:
        return torch.stack(tensors, dim=self.dim)

class Concat(torch.nn.Module):
    def __init__(self,
        dim: int=1
    ):
        super(Concat, self).__init__()
        self.dim = dim

    def forward(self, tensors: typing.List[torch.Tensor]) -> torch.Tensor:
        return torch.cat(tensors, dim=self.dim)

class Split(torch.nn.Module): #TODO: optimize by returning the tuple
    def __init__(self,
        dim: int=0, # at which dim to split
        split: int=0, # how many to split to, with 0 denoting an even split
    ):
        super(Split, self).__init__()
        self.dim = dim
        self.split = split

    def forward(self, 
        tensor: torch.Tensor,
        # index: torch.Tensor # scalar tensor denoting the split index
    ) -> typing.Dict[str, torch.Tensor]:
    # ) -> typing.Tuple[torch.Tensor, torch.Tensor]:
        ret = {}
        size = self.split if self.split else tensor.shape[self.dim] // 2
        chunks = torch.split(tensor, size, dim=self.dim) #[int(index)]
        for i in range(len(chunks)):
            ret['chunk'+str(i)] = chunks[i]
        return ret


class Slice(torch.nn.Module):
    def __init__(self,
        dim: int=0, # at which dim to slice
        start : int=0, #the index to start with
        length: int=1, #the index to end with
    ):
        super(Slice, self).__init__()
        self.dim = dim
        self.start = start
        self.length = length

    def forward(self, tensor: torch.Tensor) -> torch.Tensor:
        return torch.narrow(input = tensor,dim=self.dim,start = self.start, length = self.length)



#TODO: Is this really needed?
class SelectTensor(torch.nn.Module):
    def __init__(self):

        super(SelectTensor,self).__init__()
    
    def forward(self,
                tensors: typing.List[torch.Tensor],
                ref_tensor: torch.Tensor) -> torch.Tensor:

                for ten in tensors:
                    if ten.shape[-2:] == ref_tensor.shape[-2:]:
                        out_tensor = ten
                        break
                    else:
                       out_tensor = None
                
                return out_tensor

class Detach(torch.nn.Module):
    def __init__(self):
        super(Detach, self).__init__()

    def forward(self, tensor: torch.Tensor) -> torch.Tensor:
        return tensor.detach()

class Flatten(torch.nn.Module):
    def __init__(self):
        super(Flatten, self).__init__()
        self.flatten = torch.nn.Flatten()

    def forward(self, tensor: torch.Tensor) -> torch.Tensor:
        return self.flatten(tensor)

class ReshapeAs(torch.nn.Module):
    def __init__(self):
        super(ReshapeAs, self).__init__()

    def forward(self,
        tensor: torch.Tensor,
        shape: torch.Tensor,
    ) -> torch.Tensor:
        return tensor.reshape_as(shape)


import logging

log = logging.getLogger(__name__)

CONTEXT = None

try:
    import nvdiffrast.torch as dr    

    if CONTEXT is None:
        CONTEXT = dr.RasterizeGLContext()

    from moai.monads.render.nvdiffrast.silhouette import Silhouette as MeshSilhouette

    __all__ = [
        'MeshSilhouette',
    ]
except:
    log.error(f"The nvdiffrast package (https://github.com/NVlabs/nvdiffrast) is required to use the corresponding rendering monads. Install it by following the instructions in the respective repository.")    





from moai.monads.human.body.prior.human_body_prior import VPoser_v2
from collections import namedtuple

import torch
import typing

#NOTE: code from https://github.com/nghorbani/human_body_prior

__all__ = ["VPoser2"]

VPoser2InnerParams = namedtuple('VPoser2InnerParams', ['num_neurons', 'latentD'])
VPoser2Params = namedtuple('VPoser2Params', ['model_params'])

class VPoser2(VPoser_v2):
    def __init__(self,
        flatten_pose:       bool=True,
    ):
        super(VPoser2, self).__init__(
            VPoser2Params(VPoser2InnerParams(512, 32))
        )
        self.flatten_pose = flatten_pose
        self.eval()

    def forward(self,
        encode:         torch.Tensor=None,
        decode:         torch.Tensor=None,
        autoencode:     torch.Tensor=None,
    ) -> typing.Mapping[str, torch.Tensor]:
        out = { }        
        if autoencode is not None:
            # decoded = super(VPoser2, self).forward(autoencode)
            q_z = super(VPoser2, self).encode(autoencode)
            # q_z_sample = q_z.rsample()
            q_z_sample = q_z.mean
            decoded = super(VPoser2, self).decode(q_z_sample)
            decoded.update({
                'poZ_body_mean': q_z.mean,
                'poZ_body_std': q_z.scale,
                'q_z': q_z}
            )
            out['pose'] = decoded['pose_body']
            out['embedding'] = decoded['poZ_body_mean'] # decoded['q_z']
            if self.flatten_pose:
                out['pose'] = out['pose'].reshape(autoencode.shape[0], -1)
            return out
        if encode is not None:
            out['embedding'] = self.encode(encode)
        if decode is not None:
            out['pose'] = self.decode(decode)['pose_body']
            if self.flatten_pose:
                out['pose'] = out['pose'].reshape(decode.shape[0], -1)
        return out




from moai.utils.torch import get_submodule

import torch
import typing

__all__ = ['SupervisionParameterSelector']

class SupervisionParameterSelector(typing.Callable[[torch.nn.Module], typing.List[torch.Tensor]]):
    def __call__(self, module: torch.nn.Module) -> typing.List[torch.Tensor]:
        # return list(get_submodule(module, 'supervision').parameters())
        return { 'params': list(get_submodule(module, 'supervision').parameters()) }


from moai.monads.generation.grid.equirectangular.grid import Equirectangular as Grid

import torch

#NOTE: from WS-PSNR
'''
    double getWeight(int form,int i,int j,int width,int height){
	double a;
    //======  format0: Equirectangular  =======  
	if (form == 0) {
		a = cos((j-(height/2-0.5))*3.1415926/height);
		return a;
	}
'''

class Distortion(Grid):
    def __init__(self,
        mode:               str='latitude',
        width:              int=512,
        persistent:         bool=True,
    ):
        super(Distortion, self).__init__(
            mode='pi', order='latlong', inclusive=False,
            width=width, long_offset_pi=0.0,
            persistent=persistent,
        )
        if mode == 'latitude':
            self.grid = torch.cos(self.grid[:, 0, ...]).unsqueeze(1)



import logging
import typing

log = logging.getLogger(__name__)

__HAS_CLEAR_ML__ = False
__CLEARML_TASK__ = None

try:
    import clearml
    __HAS_CLEAR_ML__ = True
except:
    log.error(f"Please `pip install clearml` to use the ClearML engine module.")

__all__ = ['ClearML']

def _init_task(
    project_name:       str,
    task_name:          str,
    uri:                typing.Optional[str]=None,
    tags:               typing.Optional[typing.Union[str, typing.Sequence[str]]]=None,
    connect_args:       typing.Optional[typing.Union[bool, typing.Dict[str, bool]]]=True,
    connect_libs:       typing.Optional[typing.Union[bool, typing.Dict[str, bool]]]=False,
    connect_res:        typing.Optional[typing.Union[bool, typing.Dict[str, bool]]]=True,
    # connect_logs:       typing.Optional[typing.Union[bool, typing.Dict[str, bool]]]=None,
) -> None:
    global __CLEARML_TASK__
    if __CLEARML_TASK__ is None:
        __CLEARML_TASK__ = clearml.Task.init(
            project_name=project_name,
            task_name=task_name,
            task_type=clearml.TaskTypes.training,
            reuse_last_task_id=True,
            continue_last_task=False,
            output_uri=uri,
            auto_connect_arg_parser=connect_args,
            # auto_connect_frameworks=connect_libs,
            auto_connect_frameworks={
                'matplotlib': False, 'tensorflow': False,
                'tensorboard': True, 'pytorch': connect_libs,
                'xgboost': False, 'scikit': False, 'fastai': False,
                'lightgbm': False, 'hydra': True, 'detect_repository': True,
                'tfdefines': False, 'joblib': False, 'megengine': False,
                'catboost': False,
            },
            auto_resource_monitoring=connect_res,
            # auto_connect_streams=connect_logs,
            auto_connect_streams={
                'stdout': False, 'stderr': False, 'logging': True,
            },
            deferred_init=False,
        )
        if tags:
            __CLEARML_TASK__.add_tags(list(tags))
        __CLEARML_TASK__.connect_configuration("config_resolved.yaml", name='hydra')
    return __CLEARML_TASK__

def _get_logger(
    project_name:       str,
    task_name:          str,
    uri:                typing.Optional[str]=None,
    tags:               typing.Optional[typing.Union[str, typing.Sequence[str]]]=None,
) -> clearml.Logger:
    global __CLEARML_TASK__
    if __CLEARML_TASK__ is None:
        _init_task(project_name, task_name, uri, tags)
    return __CLEARML_TASK__.get_logger()

class ClearML(object):
    def __init__(self,
        project_name:       str,
        task_name:          str,
        uri:                typing.Optional[str]=None,
        tags:               typing.Optional[typing.Union[str, typing.Sequence[str]]]=None,
        connect_args:       typing.Optional[typing.Union[bool, typing.Dict[str, bool]]]=None,
        connect_libs:       typing.Optional[typing.Union[bool, typing.Dict[str, bool]]]=None,
        connect_res:        typing.Optional[typing.Union[bool, typing.Dict[str, bool]]]=None,
        # connect_logs:       typing.Optional[typing.Union[bool, typing.Dict[str, bool]]]=None,
    ) -> None:       
        _init_task(
            project_name, task_name, uri, tags, 
            connect_args, connect_libs, connect_res
        )


from moai.visualization.visdom.base import Base
from moai.utils.arguments import ensure_string_list
from moai.utils.iterators import pairwise

import torch
import visdom
import functools
import typing
import logging
import numpy as np
import cv2
import colour
import math
import toolz
from PIL import Image

log = logging.getLogger(__name__)

__all__ = ["Pose2d"]

class Pose2d(Base):
    def __init__(self,
        images:         typing.Union[str, typing.Sequence[str]],
        poses:          typing.Union[str, typing.Sequence[str]],
        gt:             typing.Union[str, typing.Sequence[str]],
        pred:           typing.Union[str, typing.Sequence[str]],
        gt_masks:       typing.Union[str, typing.Sequence[str]],
        pred_masks:     typing.Union[str, typing.Sequence[str]],
        pose_structure: typing.Union[str, typing.Sequence[str]],
        coords:         typing.Union[str, typing.Sequence[str]],
        color_gt:       typing.Union[str, typing.Sequence[str]],
        color_pred:     typing.Union[str, typing.Sequence[str]],
        name:           str="default",
        ip:             str="http://localhost",
        port:           int=8097,
        reverse_coords: bool=False,
        rotate_image:   bool=False,
        transparency:   float=0.4,
        scale:          float=1.0,
        use_mask:       bool=True,
        jpeg_quality:       int=50,
    ):
        super(Pose2d, self).__init__(name, ip, port)
        self.images = ensure_string_list(images)
        self.poses = ensure_string_list(poses)
        self.gt = ensure_string_list(gt)
        self.pred = ensure_string_list(pred)
        self.gt_masks = ensure_string_list(gt_masks)
        self.pred_masks = ensure_string_list(pred_masks)
        self.pose_structure = ensure_string_list(pose_structure)
        self.coords = ensure_string_list(coords)
        self.color_gt = list(map(colour.web2rgb, ensure_string_list(color_gt)))
        self.color_pred = list(map(colour.web2rgb, ensure_string_list(color_pred)))
        self.reverse = reverse_coords
        self.rotate = rotate_image
        self.transparency = transparency
        self.scale = scale
        self.use_mask = use_mask
        self.viz_pose = {
            'human_pose2d': functools.partial(self.__draw_human_pose2d, 
                self.visualizer, marker=cv2.MARKER_DIAMOND, 
                rotate=self.rotate, transparency=self.transparency,
                scale=self.scale, jpeg_quality=jpeg_quality
            ),
        }
        self.xforms = { #TODO: extract these into a common module
            'ndc': lambda coord, img: torch.addcmul(
                torch.scalar_tensor(0.5).to(coord), coord, torch.scalar_tensor(0.5).to(coord)
            ) * torch.Tensor([*img.shape[2:]]).to(coord).expand_as(coord),
            'coord': lambda coord, img: coord,
            'norm': lambda coord, img: coord * torch.Tensor([*img.shape[2:]]).to(coord).expand_as(coord),
        }
        self.access = lambda td, k: toolz.get_in(k.split('.'), td)

    @property
    def name(self) -> str:
        return self.env_name
        
    def __call__(self, 
        tensors:    typing.Dict[str, torch.Tensor],
        step:       typing.Optional[int]=None    
    ) -> None:
        for img, poses, gt, pred, gt_masks, pred_masks, pose_struct, gt_c, pred_c, coord in zip(
            self.images, self.poses, self.gt, self.pred, self.gt_masks, self.pred_masks,  [self.pose_structure, ] * len(self.pred),
            self.color_gt, self.color_pred, self.coords
        ):
            gt_coord = self.access(tensors, gt).detach()
            pred_coord = self.access(tensors, pred).detach()
            gt_masks = self.access(tensors, gt_masks)
            pred_masks = self.access(tensors, pred_masks)
            if self.reverse:
                gt_coord = gt_coord.flip(-1)
                pred_coord = pred_coord.flip(-1)
            image = self.access(tensors, img).detach()
            self.viz_pose[poses](
                image,
                self.xforms[coord](gt_coord, image),
                self.xforms[coord](pred_coord, image),
                gt_masks.detach() if self.use_mask else torch.ones_like(gt_coord[..., 0:1]),
                pred_masks.detach() if self.use_mask else torch.ones_like(pred_coord[..., 0:1]),
                pose_struct,
                np.uint8(np.array(list(gt_c)) * 255),
                np.uint8(np.array(list(pred_c)) * 255),
                # np.uint8(np.array(list(reversed(gt_c))) * 255),
                # np.uint8(np.array(list(reversed(pred_c))) * 255),
                coord, f"{img}_{pred}", f"{img}_{pred}", self.name
            )    
    
    @staticmethod
    def __draw_human_pose2d(
        visdom:             visdom.Visdom,
        images:             torch.Tensor,
        gt_coordinates:     torch.Tensor,
        pred_coordinates:   torch.Tensor,
        gt_masks:           torch.Tensor,
        pred_masks:         torch.Tensor,
        pose_structure:     typing.List[typing.List[int]],
        gt_color:           typing.List[float],
        pred_color:         typing.List[float],
        coord:              str,
        key:                str,
        win:                str,
        env:                str,
        marker:             int,
        rotate:             bool,
        transparency:       float,
        scale:              float,
        jpeg_quality:       int=50,
    ):
        b, _, h, w = images.shape
        imgs = np.zeros([b, 3, int(scale * h), int(scale * w)], dtype=np.uint8) if not rotate \
            else np.zeros([b, 3, int(scale * w), int(scale * h)], dtype=np.uint8)
        gt_coords = gt_coordinates.cpu().int()
        pred_coords = pred_coordinates.cpu().int()
        gt_coords = torch.flip(gt_coords, dims=[-1])
        pred_coords = torch.flip(pred_coords, dims=[-1])
        gt_coords = gt_coords.numpy()
        pred_coords = pred_coords.numpy()
        diagonal = torch.norm(torch.Tensor([*imgs.shape[2:]]), p=2)
        marker_size = int(0.015 * diagonal) #TODO: extract percentage param to config?
        line_size = int(0.005 * diagonal) #TODO: extract percentage param to config?
        line_size = int(np.clip(line_size, a_min=1, a_max=w*0.1))
        for i in range(imgs.shape[0]):
            img = images[i, ...].cpu().numpy().transpose(1, 2, 0) * 255.0
            img = img.copy().astype(np.uint8) if img.shape[2] > 1\
                else cv2.cvtColor(img.copy().astype(np.uint8), cv2.COLOR_GRAY2RGB)
            bg = img.copy()
            for coords, color, masks in zip(
                [gt_coords, pred_coords],
                [gt_color, pred_color],
                [gt_masks, pred_masks]
            ):       
                coord_i = coords[i, ...]
                for kpts_group in pose_structure:
                    for (a, b) in pairwise(kpts_group):
                    # for j in range(len(kpts_group) - 1):                    
                        if torch.sum(masks[i, a]) and torch.sum(masks[i, b]):
                            start_xy = tuple(coord_i[a])
                            end_xy = tuple(coord_i[b])
                            X = (start_xy[0], end_xy[0])
                            Y = (start_xy[1], end_xy[1])
                            mX = np.mean(X)
                            mY = np.mean(Y)
                            length = ((Y[0] - Y[1]) ** 2 + (X[0] - X[1]) ** 2) ** 0.5
                            angle = math.degrees(math.atan2(Y[0] - Y[1], X[0] - X[1]))
                            stickwidth = line_size
                            # polygon = cv2.ellipse2Poly(
                            #     (int(mX or 0), int(mY or 0)),
                            #     (int(length/2 or 1), int(stickwidth or 1)),
                            #     int(angle), 0, 360, 1
                            # )
                            # cv2.fillConvexPoly(bg, polygon, color.tolist())
                            cv2.line(img, start_xy, end_xy, color.tolist(), thickness=line_size)
                for k, coord in enumerate(coord_i):
                    if torch.sum(masks[i, k]):
                        if marker < 0:
                            cv2.circle(bg, 
                                tuple(coord), 
                                marker_size, color.tolist(), thickness=line_size
                            )
                        else:
                            cv2.drawMarker(bg, 
                                tuple(coord),
                                color.tolist(),
                                marker, marker_size, line_size
                            )

                img = cv2.addWeighted(bg, transparency, img, 1 - transparency, 0)
            h, w = img.shape[:2]
            if scale != 1.0:
                img = np.array(Image.fromarray(img).resize(
                    (int(w * scale), int(h * scale)), Image.ANTIALIAS
                ))
            imgs[i, ...] = img.transpose(2, 0, 1) if not rotate\
                else cv2.rotate(img, cv2.ROTATE_90_CLOCKWISE).transpose(2, 0, 1)
        visdom.images(
            imgs,
            # np.flip(imgs, axis=1),
            win=win,
            env=env,
            opts={
                'title': key,
                'caption': key,
                'jpgquality': jpeg_quality,
            }
        )



from moai.utils.torch import cross_product

import torch

#TODO: produces an error when using odd batch size, check it out
class NormalEstimation2d(torch.nn.Module):
    def __init__(self,
    
    ):
        super(NormalEstimation2d, self).__init__()

    def forward(self, points: torch.Tensor) -> torch.Tensor:
        v_pad = torch.nn.functional.pad(points, (0, 0, 0, 1), mode="replicate")
        u_pad = torch.cat([points, points[:, :, :, :1]], dim=-1)
        du = u_pad[:, :, :, :-1] - u_pad[:, :, :, 1:]
        dv = v_pad[:, :, :-1, :] - v_pad[:, :, 1:, :]
        normals = cross_product(dv, du, dim=1)
        return torch.nn.functional.normalize(normals)




import typing
import moai.networks.lightning as minet
import moai.utils.parsing.rtp as mirtp

import moai.networks.lightning as minet
import moai.nn.linear as mil

import torch
import inspect
import hydra.utils as hyu
import omegaconf.omegaconf
import typing
import logging
import toolz

from moai.modules.lightning.reparametrization.normal import NormalPrior

log = logging.getLogger(__name__)

__all__ = ["VariationalAutoencoder"]


class VariationalAutoencoder(minet.FeedForward):
    def __init__(self, 
        configuration:  omegaconf.DictConfig,
        io:             omegaconf.DictConfig,
        modules:        omegaconf.DictConfig,
        data:           omegaconf.DictConfig=None,
        parameters:     omegaconf.DictConfig=None,
        feedforward:    omegaconf.DictConfig=None,        
        monads:         omegaconf.DictConfig=None,        
        supervision:    omegaconf.DictConfig=None,
        validation:     omegaconf.DictConfig=None,
        visualization:  omegaconf.DictConfig=None,
        export:         omegaconf.DictConfig=None,
    ):
        super(VariationalAutoencoder, self).__init__(
            feedforward=feedforward, monads=monads, 
            visualization=visualization, data=data,
            supervision=supervision, validation=validation,
            export=export, parameters=parameters,
        )
        self.latent_dim = configuration.reparametrization.latent_dim
        self.repeat_val = configuration.repeat_val
        self.traversal_len = configuration.traversal_len
        self.traversal_step = configuration.traversal_step
        self.traversal_dim = configuration.traversal_dim
        self.traversal_init_value = configuration.traversal_init_value
        flatten = toolz.get_in(['flatten'], modules)        
        flatten = hyu.instantiate(flatten) if flatten else torch.nn.Flatten()
        prior = toolz.get_in(['reparametrization'], modules)        
        prior = hyu.instantiate(prior) if prior else NormalPrior()
        self.reparametrizer = Reparametrizer(configuration, prior, flatten) 
        self.encoder = hyu.instantiate(modules['encoder'])
        self.decoder = hyu.instantiate(modules['decoder'])
        self.enc_fwds, self.dec_fwds, self.rep_fwds = [], [], []
        hparams = {}
        hparams[f"configuration"] = configuration 
        hparams[f"io"] = io 
        hparams[f"modules"] = modules 
        self.hparams.update(hparams)

        params = inspect.signature(self.encoder.forward).parameters
        enc_in = list(zip(*[mirtp.force_list(io.encoder[prop]) for prop in params]))
        enc_out = mirtp.split_as(mirtp.resolve_io_config(io.encoder['out']), enc_in)

        self.enc_res_fill = [mirtp.get_result_fillers(self.encoder, out) for out in enc_out]        
        get_enc_filler = iter(self.enc_res_fill)
        
        for keys in enc_in:
            self.enc_fwds.append(lambda td,
                tk=keys,
                args=params.keys(),
                enc=self.encoder,
                filler=next(get_enc_filler):
                    filler(td, enc(**dict(zip(args,
                        list(
                                td[k] if type(k) is str
                                else list(td[j] for j in k)
                            for k in tk
                        )
                    ))))
            )

        params = inspect.signature(self.reparametrizer.forward).parameters
        rep_in = list(zip(*[mirtp.force_list(io.reparametrization[prop]) for prop in params]))
        rep_out = mirtp.split_as(mirtp.resolve_io_config(io.reparametrization['out']), enc_in)

        self.rep_res_fill = [mirtp.get_result_fillers(self.reparametrizer, out) for out in rep_out]        
        get_rep_filler = iter(self.rep_res_fill)
        
        for keys in rep_in:
            self.rep_fwds.append(lambda td,
                tk=keys,
                args=params.keys(),
                rep=self.reparametrizer,
                filler=next(get_rep_filler):
                    filler(td, rep(**dict(zip(args,
                        list(
                                td[k] if type(k) is str
                                else list(td[j] for j in k)
                            for k in tk
                        )
                    ))))
            ) 

        params = inspect.signature(self.decoder.forward).parameters
        dec_in = list(zip(*[mirtp.force_list(io.decoder[prop]) for prop in params]))
        dec_out = mirtp.split_as(mirtp.resolve_io_config(io.decoder['out']), dec_in)

        self.dec_res_fill = [mirtp.get_result_fillers(self.decoder, out) for out in dec_out]
        get_dec_filler = iter(self.dec_res_fill)
        
        for keys in dec_in:
            self.dec_fwds.append(lambda td,
                tk=keys,
                args=params.keys(), 
                dec=self.decoder,
                filler=next(get_dec_filler):
                    filler(td, dec(**dict(zip(args,
                        list(
                                td[k] if type(k) is str
                                else list(td[j] for j in k)
                            for k in tk
                        )
                    ))))
            )
            
    def decode(self,
        tensor: torch.Tensor
    ):
        return self.decoder(tensor)

    def forward(self, 
        td: typing.Dict[str, torch.Tensor]
    ) -> typing.Dict[str, torch.Tensor]:
        for e, r in zip(self.enc_fwds, self.rep_fwds):
            e(td)
            r(td)
        out = self.decode(td[f"z_reshaped"])
        td[f"x_hat"] = out
        return td

    def training_step_end(self, 
        train_outputs: typing.Dict[str, typing.Union[torch.Tensor, typing.Dict[str, torch.Tensor]]]
    ) -> None:
        if self.global_step and (self.global_step % self.visualization.interval == 0):
            self.visualization.visualizers[0](train_outputs['tensors'])
            self.visualization.visualizers[1](train_outputs['tensors'])
        return train_outputs['loss']



class Reparametrizer(torch.nn.Module):
    def __init__(self,
        configuration:    omegaconf.DictConfig,
        prior:            torch.nn.Module,
        flatten:          torch.nn.Module,
    )-> None:
        super(Reparametrizer, self).__init__()
        self.reparametrize = prior
        self.flatten = flatten
        self.linear_mu = mil.make_linear_block(
            block_type=configuration.linear.type,
            linear_type="linear",
            activation_type = configuration.linear.activation.type,
            in_features=configuration.reparametrization.enc_out_dim,
            out_features=configuration.reparametrization.latent_dim
        )
        self.linear_logvar = mil.make_linear_block(
            block_type=configuration.linear.type,
            linear_type="linear",
            activation_type = configuration.linear.activation.type,
            in_features=configuration.reparametrization.enc_out_dim,
            out_features=configuration.reparametrization.latent_dim
        )
        self.linear_to_dec = mil.make_linear_block(
            block_type=configuration.dec_linear.type,
            linear_type="linear",
            activation_type = configuration.linear.activation.type,
            in_features=configuration.reparametrization.latent_dim,
            out_features=configuration.reparametrization.enc_out_dim
        )

    def forward(self,
        features:   torch.Tensor,
    ) -> typing.Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        x_flat = self.flatten(features)
        mu = self.linear_mu(x_flat)
        logvar = self.linear_logvar(x_flat)
        z = self.reparametrize(mu, logvar)
        z = self.linear_to_dec(z)
        z_reshaped = z.reshape_as(features)

        return z_reshaped, mu, logvar


import moai.nn.linear as mil

import torch
import logging
import omegaconf.omegaconf

log = logging.getLogger(__name__)

__all__ = ["MLP"]

class MLP(torch.nn.Module):
    def __init__(self,
        configuration: omegaconf.DictConfig,
    ) -> None:
        super(MLP, self).__init__()
        self.block_modules = torch.nn.ModuleDict()
        in_features, mid_features = configuration.in_features, configuration.out_features
        enc_list = torch.nn.ModuleList()

        for b in range(0, configuration.blocks):
            enc_list.append(mil.make_linear_block(
                block_type=configuration.linear.type,
                linear_type="linear",
                in_features=in_features if b == 0  else mid_features,
                out_features=mid_features,
                activation_type=configuration.linear.activation.type
            ))
        enc_list.append(mil.make_linear_block(
                block_type=configuration.prediction.type,
                linear_type="linear",
                in_features=mid_features,
                out_features=mid_features,
                activation_type=configuration.prediction.activation.type
            ))
        self.sequential = torch.nn.Sequential(*enc_list)

    def forward(self,
        x:  torch.Tensor,
    ) -> torch.Tensor:
        return self.sequential(x)


from moai.utils.arguments import assert_path
from scipy.spatial.transform import Rotation as R

import pickle
import toolz
import numpy as np
import glob
import torch
import os
import typing
import logging
import tqdm
import smplx

log = logging.getLogger(__name__)

__all__ = ['THuman2']

class THuman2(torch.utils.data.Dataset):
    def __init__(self,
        data_root:          str,
        smplx_root:         str,
        reconstruct:        bool=False,
    ) -> None:
        super().__init__()
        assert_path(log, 'THuman2.0 data root path', data_root)
        assert_path(log, 'SMPLX data path', smplx_root)
        self.data = []
        self.reconstruct = reconstruct
        for sample in tqdm.tqdm(glob.glob(os.path.join(data_root, 'smplx', '**', '*.pkl')), desc='Loading THuman2.0 data'):
            with open(sample, 'rb') as f:
                data = pickle.load(f)
            self.data.append(toolz.valmap(
                lambda a: torch.from_numpy(a.squeeze()).float(), data
            ))
        if self.reconstruct:
            self.body = smplx.create(
                model_path=smplx_root, model_type='smplx',
                num_expression_coeffs=10, create_expression=True,
                create_jaw_pose=False, create_leye_pose=False,
                create_reye_pose=False, use_face_contour=False,
                batch_size=1, age='adult', create_left_hand_pose=False,
                create_right_hand_pose=False, use_pca=True,
                num_pca_comps=12, gender='neutral',
                flat_hand_mean=False, create_betas=False,
                num_betas=10, create_global_orient=False,
                create_body_pose=False, create_transl=False,
            )
            self.body.requires_grad_(False)
        log.info(f"Loaded {len(self)} THuman2.0 samples.")

    def __len__(self) -> int:
        return len(self.data)

    def __getitem__(self, index: int) -> typing.Dict[str, torch.Tensor]:
        item = self.data[index]
        #fix global rotation
        r =  R.from_euler('xz', [90,-90], degrees=True)
        new_global_orientation = R.from_matrix(
            (r.as_matrix() @ R.from_rotvec(item['global_orient']).as_matrix())
        ).as_rotvec()
        out = {
            'smplx': {
                'gender': 0,
                'scale': item['scale'], 
                'params': {                
                    'betas': item['betas'],
                    'expression': item['expression'],
                    'body_pose': item['body_pose'],
                    'transl': item['translation'],
                    'global_orient': torch.from_numpy(new_global_orientation).float(),
                    'left_hand_pose': item['left_hand_pose'],
                    'right_hand_pose': item['right_hand_pose'],
                    'jaw_pose': item['jaw_pose'],
                    'left_eye_pose': item['leye_pose'],
                    'right_eye_pose': item['reye_pose'],
                },
            },
            '__moai__': type(self).__name__
        }
        if self.reconstruct:
            with torch.no_grad():
                body = self.body.forward(
                    global_orient=torch.from_numpy(new_global_orientation).float()[np.newaxis, ...],
                    betas=out['smplx']['params']['betas'][np.newaxis, ...],
                    body_pose=out['smplx']['params']['body_pose'][np.newaxis, ...],
                    left_hand_pose=out['smplx']['params']['left_hand_pose'][np.newaxis, ...],
                    right_hand_pose=out['smplx']['params']['right_hand_pose'][np.newaxis, ...],
                    jaw_pose=out['smplx']['params']['jaw_pose'][np.newaxis, ...],
                    leye_pose=out['smplx']['params']['left_eye_pose'][np.newaxis, ...],
                    reye_pose=out['smplx']['params']['right_eye_pose'][np.newaxis, ...],
                    transl=out['smplx']['params']['transl'][np.newaxis, ...],
                    expression=out['smplx']['params']['expression'][np.newaxis, ...],
                )
                out['smplx'].update({
                    'mesh': {
                        'vertices': body.vertices[0],                
                        'faces': self.body.faces_tensor,
                        },
                    'joints': body.joints[0, ...],
                })
        return out


import torchvision
import torch
import hydra.utils as hyu
import omegaconf.omegaconf
import typing
import logging

log = logging.getLogger(__name__)

__all__ = ["Combine"]

class Combine(torch.utils.data.Dataset):
    def __init__(self,
        dataset:        torch.utils.data.Dataset,        
        inputs:         typing.Sequence[str],
        outputs:        typing.Sequence[str],
        transforms:     omegaconf.DictConfig={},
    ):
        super(Combine, self).__init__()
        self.inner, self.inputs, self.outputs = dataset, inputs, outputs
        self.transforms = []
        log.warning("Using an ordered composition of the following torchvision transforms:")
        for i, (k, v) in enumerate(transforms.items()):
            xform = hyu.instantiate(v)
            log.info(f"\t\t\t{i + 1}. A `{k}` transform.")
            self.transforms.append(xform)
        self.composition = torchvision.transforms.Compose(self.transforms)

    def __len__(self) -> int:
        return len(self.inner)

    def __getitem__(self, index: int) -> typing.Dict[str, torch.Tensor]:
        #NOTE: tensor/PIL image conversions are expected to be handled via the transforms order
        item = self.inner[index]
        for i, o in zip(self.inputs, self.outputs):            
            item[o] = self.composition(item[i])
        return item


from moai.utils.torch import get_submodule

import hydra
import sys
import omegaconf.omegaconf
import typing
import logging
import pytorch_lightning as pl
import torch
import os

log = logging.getLogger(__name__)

try:
    import blobconverter
except:
    log.warning("The `blobconverter` package was not found, exporting to OpenVivo is disabled.")

class TraceWrapper(pl.LightningModule):
    def __init__(self,
        module_names: typing.List[str], #exported module names
        output_names: typing.List[str], #output names
        input_names: typing.List[str], #input names
        model: torch.nn.Module,
    ) -> None:
        
        super().__init__()

        
        self.model = model
        self.exported_module_names = module_names
        self.input_names = input_names
        self.output_names = output_names

    
    def forward(self,
        x: typing.Dict[str, torch.Tensor],
    ) -> typing.Dict[str, torch.Tensor]:
        for module in self.exported_module_names:
            try:
                m = get_submodule(self, module) if module != "model" \
                    else self.model
            except:
                log.error(f'The {m} module does not exist in your model!')
                log.info(f'Please select on of the following {*self.model._modules.keys(),}')
            x = m(x)
        return {out_key: x[out_key] for out_key in self.output_names}




def export(cfg):
    hydra.utils.log.debug(f"Configuration:\n{omegaconf.OmegaConf.to_yaml(cfg, resolve=True)}")
    with open("config_resolved.yaml", 'w') as f:
        f.write(omegaconf.OmegaConf.to_yaml(cfg, resolve=True, sort_keys=False))
    with open("config.yaml", 'w') as f:
        f.write(omegaconf.OmegaConf.to_yaml(cfg, resolve=False, sort_keys=False))
    engine = hydra.utils.instantiate(cfg.engine)
    model = hydra.utils.instantiate(cfg.model)
    #TODO: add cpu device support
    device = cfg.tester.gpus[0] if 'tester' in cfg.keys() \
        else cfg.trainer.gpus[0]

    model.eval()
    model.initialize_parameters()
    model.to(device)


    trwrapper = TraceWrapper(
        cfg.export.module_names,
        cfg.export.output_names,
        cfg.export.input_names,
        model
    )

    trwrapper.eval()
    trwrapper.to(device)
    
    input_dict = {}

    for in_name in cfg.export.input_names:
        input_dict[in_name] = torch.randn(tuple(cfg.export.input_tensor[in_name])).to(device).float()

    if cfg.export.mode == 'onnx':
        log.info("exporting model to onnx!")
        trwrapper.to_onnx(
            os.path.join(cfg.export.output_path, f'{cfg.export.name}.onnx'),
            # {trwrapper.input_names[0]: input_sample},
            input_dict, 
            export_params=cfg.export.export_params, #bool 
            opset_version=cfg.export.opset_version, #int; default is 12
            input_names=trwrapper.input_names,
            output_names=trwrapper.output_names,
        )
    
    elif cfg.export.mode == 'jit':
        #TODO: Add trace support
        #trace
        log.info("converting model to be jit compatible!")
        traced_model = torch.jit.trace(
            trwrapper, #model 
            {trwrapper.input_names[0]: input_sample}, #example inputs
            strict=False, # to allow mutable (e.g. dicts) output
            )
        # traced_model = torch.jit.trace(
        #     t, #model 
        #     {t.model_in[0][0]: input_sample}, #example inputs
        #     strict=False, # to allow mutable (e.g. dicts) output
        #     )
        traced_model.save(
            os.path.join(cfg.export.output_path, f'{cfg.export.name}.pt')
        ) # Save
    
    elif cfg.export.mode == 'blob':
        #first convert to onnx
        log.info("exporting model to onnx!")
        trwrapper.to_onnx(
            os.path.join(cfg.export.output_path, f'{cfg.export.name}.onnx'),
            # {trwrapper.input_names[0]: input_sample},
            input_dict, 
            export_params=cfg.export.export_params, #bool 
            opset_version=cfg.export.opset_version, #int; default is 12
            input_names=trwrapper.input_names,
            output_names=trwrapper.output_names,
            verbose=True,
        )
        blob_path=blobconverter.from_onnx(
            model=os.path.join(cfg.export.output_path, f'{cfg.export.name}.onnx'),
            data_type=cfg.export.data_type,
            shaves=cfg.export.shaves,
            output_dir=cfg.export.output_path
        )
        #delete onnx file?


    log.info(f"Model has been saved in {cfg.export.output_path}!")


if __name__ == "__main__":  
    config_filename = sys.argv.pop(1) #TODO: argparser integration?
    sys.argv.append("hydra.run.dir=actions/${hydra.job.name}/${now:%Y-%m-%d}/${now:%H-%M-%S}-${export.name}")
    export = hydra.main(config_path="conf", config_name=config_filename)(export)
    export()


from moai.utils.arguments import assert_path
from collections import OrderedDict

import functools
import toolz
import numpy as np
import glob
import torch
import os
import typing
import logging
import bisect
import smplx

log = logging.getLogger(__name__)

__all__ = ['AMASS']

class AMASS(torch.utils.data.Dataset):
    def __init__(self,
        data_root:          str,
        smplx_root:         str='',
        device:            list=[-1],
        parts:              typing.Union[typing.List[str], str]='**',
        downsample_factor:  int=1,
        reconstruct:        bool=False,      
    ) -> None:
        super().__init__()
        assert_path(log, 'AMASS data root path', data_root)
        assert_path(log, 'SMPLX data path', smplx_root)
        is_all_parts = isinstance(parts, str) and ('all' == parts or '**' == parts)
        parts = os.listdir(data_root) if is_all_parts else parts
        parts = [parts] if isinstance(parts, str) else parts
        self.items = OrderedDict()
        self.sampling = downsample_factor
        frame_counter = 0
        self.subjects = { }
        self.device = device[0] if device[0] >= 0 else 'cpu'
        for part in parts:            
            for subject in filter(os.path.isdir, glob.glob(os.path.join(data_root, part, '**'))):
                subject_name = os.path.basename(subject)                
                gendered_shape_fn = toolz.first(glob.glob(
                    os.path.join(data_root, part, subject_name, '*_stagei.npz')
                ))
                gender = os.path.basename(gendered_shape_fn).split('_')[0]
                shape = torch.from_numpy(np.load(gendered_shape_fn, allow_pickle=False)['betas']).float().clone()
                self.subjects[f"{part}_{subject_name}"] = shape
                for action_fn in glob.glob(os.path.join(data_root, part, subject, '*_stageii.npz')):
                    data = np.load(action_fn, allow_pickle=False)
                    #frame_counter += int(data['trans'].shape[0] / downsample_factor)
                    frame_counter += max(1,int(data['trans'].shape[0] / downsample_factor))
                    self.items[frame_counter] = { 
                        'data': data, # action_fn, # data, 
                        'gender': gender,
                        'subject': f"{part}_{subject_name}",
                    }
                    del data
        self.frame_keys = list(self.items.keys())
        self.reconstruct = reconstruct
        if self.reconstruct:
            with torch.no_grad():
                smplx_create = functools.partial(smplx.create, 
                    model_path=smplx_root, model_type='smplx',
                    num_expression_coeffs=10, create_expression=True,
                    create_jaw_pose=False, create_leye_pose=False,
                    create_reye_pose=False, use_face_contour=False,
                    batch_size=1, age='adult', create_left_hand_pose=False,
                    create_right_hand_pose=False, use_pca=False,            
                    flat_hand_mean=False, create_betas=False,
                    num_betas=16, create_global_orient=False,
                    create_body_pose=False, create_transl=False,
                )
                self.bodies = {
                    'female': smplx_create(gender='female').to(self.device),
                    'male': smplx_create(gender='male').to(self.device),
                    'neutral': smplx_create(gender='neutral').to(self.device),
                }
                del smplx_create
            for v in self.bodies.values():
                v.requires_grad_(False)
        log.info(f"Loaded {len(self)} AMASS samples.")

    def __len__(self) -> int:
        return toolz.last(self.items)

    def __getitem__(self, index: int) -> typing.Dict[str, torch.Tensor]:
        ind = bisect.bisect_left(self.frame_keys, index)
        ind += (self.frame_keys[ind] == index)
        frame = index - (self.frame_keys[ind-1] if ind > 0 else 0)
        frame *= self.sampling
        item = self.items[self.frame_keys[ind]]
        gender = item['gender']
        betas = self.subjects[item['subject']]
        data = item['data']
        #frame = min(frame,(len(data['pose_body'])-1))
        # data = np.load(item['data'], allow_pickle=False)
        lhand, rhand = torch.from_numpy(data['pose_hand']).split(45, dim=1)
        leye, reye = torch.from_numpy(data['pose_eye']).split(3, dim=1)        
        out = {
            'smplx': {
                'gender': 1 if gender == 'male' else 2,
                'scale': torch.scalar_tensor(1.0), 
                'params': {                
                    'betas': betas,
                    'body_pose': torch.from_numpy(data['pose_body'][frame]).float().clone(), # 63
                    'transl': torch.from_numpy(data['trans'][frame]).float().clone(), # 3
                    'global_orient': torch.from_numpy(data['root_orient'][frame]).float().clone(), # 3
                    'left_hand_pose': lhand[frame].float().clone(), # 45
                    'right_hand_pose': rhand[frame].float().clone(), # 45
                    'jaw_pose': torch.from_numpy(data['pose_jaw'][frame]).float().clone(), # 3
                    'left_eye_pose': leye[frame].float().clone(), # 3
                    'right_eye_pose': reye[frame].float().clone(), # 3
                },
            },
            '__moai__': type(self).__name__
        }
        if self.reconstruct:
            body = self.bodies[gender].forward(
                global_orient=out['smplx']['params']['global_orient'][np.newaxis, ...].to(self.device),
                betas=out['smplx']['params']['betas'][np.newaxis, ...].to(self.device),
                body_pose=out['smplx']['params']['body_pose'][np.newaxis, ...].to(self.device),
                left_hand_pose=out['smplx']['params']['left_hand_pose'][np.newaxis, ...].to(self.device),
                right_hand_pose=out['smplx']['params']['right_hand_pose'][np.newaxis, ...].to(self.device),
                jaw_pose=out['smplx']['params']['jaw_pose'][np.newaxis, ...].to(self.device),
                leye_pose=out['smplx']['params']['left_eye_pose'][np.newaxis, ...].to(self.device),
                reye_pose=out['smplx']['params']['right_eye_pose'][np.newaxis, ...].to(self.device),
                transl=out['smplx']['params']['transl'][np.newaxis, ...].to(self.device),
            )
            out['smplx'].update({
                'mesh': {
                    'vertices': body.vertices[0].cpu(),                
                    'faces': self.bodies[gender].faces_tensor.cpu(),
                    },
                'joints': body.joints[0].cpu(),
            })
        return out


from moai.monads.masking.binary import Binary
from moai.monads.masking.finite import Finite
from moai.monads.masking.mask import (
    Mask, 
    Index,
)
from moai.monads.masking.threshold import (
    Threshold,
    LowerThan,
    HigherThan,
)

__all__ = [    
    "Binary",
    "Finite",
    "Mask",
    "Threshold",
    "LowerThan",
    "HigherThan",
    "Index",
]


from moai.data.iterator.indexed import Indexed
from moai.data.iterator.zip import (
    Zipped,
    SubsetZipped,
)
from moai.data.iterator.concat import Concatenated
from moai.data.iterator.interleave import Interleaved

__all__ = [
    "Indexed",
    "Zipped",
    "SubsetZipped",
    "Concatenated",
    "Interleaved",
]


import hydra
import sys
import omegaconf.omegaconf
import logging
import typing
from omegaconf import OmegaConf
import re
import os
import hydra.utils as hyu
from moai.engine.callbacks import ModelCallbacks


log = logging.getLogger(__name__)

def get_conf(path: str) -> omegaconf.OmegaConf:
    if os.path.splitext(path)[1] != '.yaml':
        nominal_config_path = os.path.join(path, 'config.yaml')
        if os.path.exists(nominal_config_path):
            path = nominal_config_path
        elif os.path.basename(path) != '.hydra':
            path = os.path.join(path, '.hydra') #TODO: merge with overrides
            path = os.path.join(path, 'config.yaml')
    return path

def assign(cfg: omegaconf.DictConfig, attr: str) -> typing.Union[typing.Any,typing.Any]:
    return getattr(cfg, attr) if hasattr(cfg, attr) else None

@hydra.main(config_name="tools/resume.yaml", config_path="conf")
def resume(cfg):
    #find last checkpoint
    regex = re.compile('(last.*ckpt)')
    last_ckpt_path = None
    for root, dirs, files in os.walk(os.getcwd()):
        for file in files:
            if regex.match(file):
                last_ckpt_path = os.path.join(root,file)
                break
    if not last_ckpt_path:
        log.error(f"The provided experiment path {cfg.experiment.path} does not contain a last checkpoint file.")
    
    resume_callback = hyu.instantiate(
        {
            "_target_": "moai.parameters.initialization.lightning.Continued",
            "filename": last_ckpt_path
        })
    config_file = get_conf(os.getcwd())
    with open(config_file, 'r') as f:
        init_cfg = OmegaConf.load(f)
    omegaconf.OmegaConf.set_struct(init_cfg, True)
    cfg = OmegaConf.merge(init_cfg,cfg) #to change param values
    hydra.utils.log.debug(f"Configuration:\n{omegaconf.OmegaConf.to_yaml(cfg, resolve=True)}")
    engine = hydra.utils.instantiate(cfg.engine)
    model = hydra.utils.instantiate(cfg.model,
        data=cfg.data,
        visualization=assign(cfg, "visualization"),
        export=assign(cfg, "export"),    
    )
    for name, remodel in (assign(cfg, "remodel") or {}).items():
        hydra.utils.instantiate(remodel)(model)
    model_callbacks = ModelCallbacks(model=model)
    model_callbacks += [resume_callback]
    trainer = hydra.utils.instantiate(cfg.trainer, 
        logging=assign(cfg, "logging"),
        model_callbacks=model_callbacks,
    )
    log.info("Training started.")     
    trainer.run(model)
    log.info("Training completed.")

if __name__ == "__main__":
    # os.environ['HYDRA_FULL_ERROR'] = '1'
    config_filename = sys.argv.pop(1) #TODO: argparser integration?
    sys.argv.append("hydra.run.dir=actions/${hydra.job.name}/${now:%Y-%m-%d}/${now:%H-%M-%S}-${experiment.name}")    
    resume = hydra.main(config_path="conf", config_name=config_filename)(resume)
    resume()


from moai.supervision.losses.regression.balanced.mse import ScalarMSE as BalancedScalarMSE
from moai.supervision.losses.regression.balanced.mse import VectorMSE as BalancedVectorMSE

__all__ = [
    'BalancedScalarMSE',
    'BalancedVectorMSE',
]


import torch

class DepthFromStereo(torch.nn.Module):
    def __init__(self,
        baseline:   float,          # physical distance between cameras in mm
    ):
        super(DepthFromStereo, self).__init__()
        self.base = baseline

    def forward(self,
        im_disp:    torch.Tensor,
        intrinsics:  torch.Tensor,
    ) -> torch.Tensor:
        if len(im_disp.shape) < 4:
            im_disp = im_disp.unsqueeze(0).clone()
            
        return self.base * intrinsics[0][0][0] / im_disp




from moai.data.datasets.common.image2d import (
    load_color_image,
    load_gray_image,
    load_depth_image,
    load_mask_image,
    load_normal_image,
)
from moai.data.datasets.common.raw import (
    load_npz_file,
    load_pkl_file,
)

__all__ = [
    "load_color_image",
    "load_gray_image",
    "load_depth_image",
    "load_normal_image",
    "load_mask_image",
    "load_npz_file",
    "load_pkl_file",
]


from moai.utils.arguments import assert_path

import functools
import toolz
import numpy as np
import glob
import torch
import os
import typing
import logging
import tqdm
import smplx
import orjson

log = logging.getLogger(__name__)

__all__ = ['Fit3D']

# try:
#     import orjson
#     @lru_cache(maxsize=None)
#     #@functools.lru_cache(maxsize=None)
#     def _load_json(filename: str) -> dict:
#         with open(filename, 'rb') as f:
#             data = orjson.loads(f.read())
#         return data
# except:
#     log.warning("Could not load the `orjson` package which will improve loading speeds, please consider `pip install orjson`.")
#     import json
#     @functools.lru_cache(maxsize=None)
#     def _load_json(filename: str) -> dict:
#         with open(filename, 'rb') as f:
#             data = json.load(f)
#         return data

class Fit3D(torch.utils.data.Dataset):
    def __init__(self,
        data_root:          str,
        smplx_root:         str,
        subjects:           typing.Union[typing.List[str], str]='**',
        actions:            typing.Union[typing.List[str], str]='**',
        downsample_factor:  int=1,
        device:             typing.Union[int, typing.Sequence[int]]=[-1],
        reconstruct:        bool=False,
    ) -> None:
        super().__init__()
        assert_path(log, 'Fit3D data root path', data_root)
        assert_path(log, 'SMPLX data path', smplx_root)
        is_all_subjects = isinstance(subjects, str) and ('all' == subjects or '**' == subjects)
        subjects = os.listdir(data_root) if is_all_subjects else subjects
        subjects = [subjects] if isinstance(subjects, str) else subjects
        is_all_actions = isinstance(actions, str) and ('all' == actions or '**' == actions)
        actions = list(map(lambda fn: os.path.splitext(fn)[0], 
            toolz.unique(map(
                os.path.basename, 
                glob.glob(os.path.join(data_root, '**', 'smplx', '*.json'))
            ))
        )) if is_all_actions else actions
        actions = [actions] if isinstance(actions, str) else actions
        self.device = f"cuda:{device[0]}" if isinstance(device, typing.Sequence) and device[0] >= 0 else 'cpu'
        self.data = {
            'transl': torch.empty((0, 3)),
            'global_orient': torch.empty((0, 9)),
            'body_pose': torch.empty((0, 21 * 9)),
            'betas': torch.empty((0, 10)),
            'left_hand_pose': torch.empty((0, 15 * 9)),
            'right_hand_pose': torch.empty((0, 15 * 9)),
            'jaw_pose': torch.empty((0, 9)),
            'leye_pose': torch.empty((0, 9)),
            'reye_pose': torch.empty((0, 9)),
            'expression': torch.empty((0, 10)),
         }
        self._load_json = functools.lru_cache()(self.__load_json)
        for subj in tqdm.tqdm(subjects, desc='Loading Fit3D subjects'):                        
            for action_fn in tqdm.tqdm(
                glob.glob(os.path.join(data_root, subj, 'smplx', "*.json")),
                desc='Loading Fit3D subject actions'
            ):
                if os.path.splitext(os.path.basename(action_fn))[0] not in actions:
                    continue
                data = self._load_json(action_fn)
                for k, v in data.items():
                    self.data[k] = torch.cat([
                        self.data[k], 
                        torch.from_numpy(np.array(v, dtype=np.float32).squeeze())[::downsample_factor, ...].flatten(1)
                    ])                                        
        self.reconstruct = reconstruct
        if self.reconstruct:
            self.body = smplx.SMPLXLayer(
                model_path=os.path.join(smplx_root, 'smplx'), model_type='smplx',
                num_expression_coeffs=10, batch_size=1, age='adult', 
                use_pca=False, flat_hand_mean=False,
                num_betas=10, gender='neutral'
            ).to(self.device)
            self.body.requires_grad_(False)
        log.info(f"Loaded {len(self)} Fit3D samples.")

    def __load_json(self,filename: str) -> dict:
        with open(filename, 'rb') as f:
            data = orjson.loads(f.read())
        return data

    def __len__(self) -> int:
        return len(self.data['betas'])

    def __getitem__(self, index: int) -> typing.Dict[str, torch.Tensor]:
        item = toolz.itemmap(lambda d: (d[0], d[1][index]), self.data)
        out = {
            'smplx': {
                'gender': 0,
                'scale': torch.scalar_tensor(1.0), 
                'params': {                
                    'betas': item['betas'],
                    'body_pose': item['body_pose'],
                    'transl': item['transl'],
                    'global_orient': item['global_orient'],
                    'left_hand_pose': item['left_hand_pose'],
                    'right_hand_pose': item['right_hand_pose'],
                    'jaw_pose': item['jaw_pose'],
                    'left_eye_pose': item['leye_pose'],
                    'right_eye_pose': item['reye_pose'],
                }
            },
            '__moai__': type(self).__name__
        }
        if self.reconstruct:
            with torch.no_grad():
                body = self.body.forward(
                    global_orient=out['smplx']['params']['global_orient'][np.newaxis, ...].to(self.device),
                    betas=out['smplx']['params']['betas'][np.newaxis, ...].to(self.device),
                    body_pose=out['smplx']['params']['body_pose'][np.newaxis, ...].to(self.device),
                    left_hand_pose=out['smplx']['params']['left_hand_pose'][np.newaxis, ...].to(self.device),
                    right_hand_pose=out['smplx']['params']['right_hand_pose'][np.newaxis, ...].to(self.device),
                    jaw_pose=out['smplx']['params']['jaw_pose'][np.newaxis, ...].to(self.device),
                    leye_pose=out['smplx']['params']['left_eye_pose'][np.newaxis, ...].to(self.device),
                    reye_pose=out['smplx']['params']['right_eye_pose'][np.newaxis, ...].to(self.device),
                    transl=out['smplx']['params']['transl'][np.newaxis, ...].to(self.device),
                    pose2rot=False,
                )
                out['smplx'].update({
                    'mesh': {
                        'vertices': body.vertices[0].cpu(),                
                        'faces': self.body.faces_tensor.cpu(),
                        },                
                    'joints': body.joints[0, ...].cpu(),

                })
        return out


import pickle
import torch
import os
import numpy as np
import logging

log = logging.getLogger(__name__)

__all__ = ['BodyTransfer']

class BodyTransfer(torch.nn.Module):
    def __init__(self,
        weights_path:   str,
        persistence:    bool=True,
    ) -> None:
        super().__init__()
        name, ext = os.path.split(weights_path)
        if ext == '.pkl':
            with open(weights_path, 'rb') as f:
                w = pickle.load(f, encoding='latin1')['mtx']
                if hasattr(w, 'todense'):
                    w = w.todense()
                w = w[:, :w.shape[-1] // 2]
        elif ext == '.npz' or ext == '.npy':
            w = np.load(weights_path)
        else:
            log.error(f"Unsupported extension ({ext}) to load body tranfer weights from when loading {weights_path}. Can only load *.pkl, *.npy, *.npz files")
        self.register_buffer('weights', torch.from_numpy(w).float(), persistent=persistence)

    def forward(self, vertices: torch.Tensor) -> torch.Tensor:
        return torch.einsum('bni,mn->bmi', vertices, self.weights)


from moai.utils.torch import get_submodule
from moai.networks.lightning import FeedForward
from pytorch_lightning.callbacks import Callback
from collections import OrderedDict

import pprint
import omegaconf.omegaconf
import sys
import toolz
import logging

log = logging.getLogger(__name__)

__all__ = ["LossSchedule"]

class LossSchedule(Callback):
    def __init__(self,        
        milestones: omegaconf.DictConfig,        
    ):
        super().__init__()
        self.milestones = OrderedDict(sorted(
            toolz.keymap(int, omegaconf.OmegaConf.to_container(milestones)).items()
        ))
        desc = pprint.pformat(dict(toolz.keymap(lambda x: f"epoch_{x}",omegaconf.OmegaConf.to_container(milestones)).items()))
        log.info(f"A weighting supervision schedule is used:\n{desc}.")

    def on_train_epoch_start(self, trainer, pl_module: FeedForward):
        next = int(toolz.peek(self.milestones)[0]) if len(self.milestones) else sys.maxsize
        supervision = get_submodule(pl_module, 'supervision')
        if trainer.current_epoch + 1 >= next:
            ms = self.milestones.pop(next)
            for k, v in ms.items():
                supervision.weights[k] = v
                log.info(f"Updated weight ({v}) for loss [{k}] @ epoch {next}.")


from moai.engine.modules.clearml import _get_logger

import numpy as np
import pytorch_lightning
import logging
import typing
import toolz

log = logging.getLogger(__name__)

__all__ = ["ClearML"]

class ClearML(pytorch_lightning.loggers.base.LightningLoggerBase):
    def __init__(self,
        project_name:   str,
        task_name:      str,
        uri:            typing.Optional[str]=None,
        tags:           typing.Optional[typing.Union[str, typing.Sequence[str]]]=None,
    ):
        super(ClearML, self).__init__()
        self._name, self._version = project_name, task_name
        self.logger = _get_logger(project_name, task_name, uri, tags)
        # self.keys = {
        #     'train': 'train_loss', 'val': 'val_metrics', 'test': 'test_results',
        #     'epoch': 'epoch', 'step': 'global_step'
        # }

    @pytorch_lightning.loggers.base.rank_zero_only
    def log_metrics(self, 
        metrics:        typing.Dict[str, typing.Any],
        step:           int
    ) -> None:
        train_metrics = toolz.keymap(lambda k: k.replace('train_', ''), 
            toolz.keyfilter(lambda k: k.startswith('train_'), metrics)
        )
        val_metrics = toolz.keymap(lambda k: k.replace('val_', ''), 
            toolz.keyfilter(lambda k: k.startswith('val_'), metrics)
        )
        test_metrics = toolz.keymap(lambda k: k.replace('test_', '').replace('/epoch_0', ''), 
            toolz.keyfilter(lambda k: k.startswith('test_'), metrics)
        )
        e = int(metrics['epoch'])
        if train_metrics:            
            loss = float(metrics['total_loss'])
            self.logger.report_scalar('train', 'loss', loss, step)
            for k, v in train_metrics.items():
                self.logger.report_scalar('train', k, v, step)
        elif test_metrics:
            return #TODO: test case 
        if val_metrics:            
            dataset_val_metrics = toolz.valmap(
                lambda v: toolz.keymap(lambda k: k.split('/')[0], dict(v)), 
                toolz.groupby(
                    lambda k: toolz.get(1, k[0].split('/'), 'metrics'),
                    val_metrics.items()
                )
            )
            for d, m in dataset_val_metrics.items():
                for k, v in m.items():
                    self.logger.report_scalar(d, k, v, e)

    @pytorch_lightning.loggers.base.rank_zero_only
    def log_hyperparams(self,
        params: typing.Dict[str, typing.Any] #TODO: or namespace object ?
    ) -> None:
        pass    

    @pytorch_lightning.loggers.base.rank_zero_only
    def save(self) -> None:
        pass

    @pytorch_lightning.loggers.base.rank_zero_only
    def finalize(self, 
        status: str
    ) -> None:
        """Do any processing that is necessary to finalize an experiment
            :param status: Status that the experiment finished with (e.g. success, failed, aborted)
        """
        self.save()    

    @property
    def name(self) -> str:
        return self._name

    @property
    def version(self) -> int:
        return self._version

    @property
    def experiment(self) -> typing.Any:
        return self.name


from moai.nn.sqex import SqueezeExcite as SE

import torch
import omegaconf.omegaconf

__all__ = ['SqueezeExcite']

class SqueezeExcite(torch.nn.Module):
    def __init__(self,
        features:       int,
        squeeze:        omegaconf.DictConfig, # type: 'averageXd', # one of ['averageXd', 'attentionXd']
        excite:         omegaconf.DictConfig, # delta: activation_{type|params}, # ratio \in [0, 1], # operation_{type|params}: one of ['convXd', 'linear']
        mode:           str='mul',
        expansion:      int=2,
    ):
        super(SqueezeExcite, self).__init__()
        self.se = SE(
            features=features,
            squeeze=squeeze,
            excite=excite,
            mode=mode,
        )

    def forward(self, 
        enc:    torch.Tensor,
        dec:    torch.Tensor,
        gate:   torch.Tensor=None,
    ) -> torch.Tensor:
        se = self.se(enc)
        return torch.cat([dec, se], dim=1)


from moai.utils.arguments import assert_positive

from pytorch3d.structures import Meshes
from pytorch3d.renderer import (
    PerspectiveCameras,    
    RasterizationSettings,
    MeshRenderer,
    MeshRasterizer,
    BlendParams,
    SoftSilhouetteShader,
    TexturesVertex,
)

import torch
import logging
import numpy as np
import typing

__all__ = ['Silhouette']

log = logging.getLogger(__name__)

class Silhouette(torch.nn.Module):
    def __init__(self,        
        focal_length:       typing.Union[float, typing.Tuple[float, float]]=5000.0,
        principal_point:    typing.Union[float, typing.Tuple[float, float]]=0.5,
        width:              int=None,
        height:             int=None,
        rotation:           torch.Tensor=None,
        translation:        torch.Tensor=None,
        z_reflection:       bool=True,
        faces_per_pixel:    int=1,
        blend_sigma:        float=1e-7,
        blend_gamma:        float=1e-7,
        raster_bin_size:    typing.Optional[int]=None,
        raster_blur_radius: float=1e-4,
        persistent:         bool=False,
        max_faces_per_bin:  typing.Optional[int]=None,
    ) -> None:
        super().__init__()
        self.focal_length = (focal_length, focal_length) if isinstance(focal_length, float) else focal_length
        self.principal_point = (principal_point, principal_point) if isinstance(principal_point, float) else principal_point
        self.resolution = (width, height)
        if principal_point is None:
            assert_positive(log, "width", width)
            assert_positive(log, "height", height)
            self.principal_point = (float(width // 2), float(height // 2))
            self.has_nominal_principal = True
        else:
            self.has_nominal_principal = False
        self.register_buffer('adjust_translation', 
            torch.Tensor([[1.0, 1.0, -1.0]]).float() if z_reflection else
            torch.ones(1, 3), persistent=persistent
        )
        if rotation is None:
             rotation = torch.eye(3)[np.newaxis, ...].float()
        rotation = torch.nn.Parameter(rotation, requires_grad=True)
        self.register_parameter('rotation', rotation)
        if translation is None:
            translation = torch.zeros([1, 3]).float()
        translation = torch.nn.Parameter(translation, requires_grad=True)
        self.register_parameter('translation', translation)
        self.blend_params = BlendParams(sigma=blend_sigma, gamma=blend_gamma)
        self.faces_per_pixel = faces_per_pixel
        self.raster_blur_radius = raster_blur_radius
        self.raster_bin_size = raster_bin_size
        self.max_faces_per_bin = max_faces_per_bin

    def forward(self,
        vertices:           torch.Tensor, # [B, V, 3]
        faces:              torch.Tensor, # [B, F, 3]
        rotation:           torch.Tensor, # [B, 3, 3]
        translation:        torch.Tensor, # [B, 3]
        image:              torch.Tensor=None,
        nominal_image:      torch.Tensor=None,
    ):
        b = vertices.shape[0]
        T = translation if translation is not None else self.translation.expand(b, 3)
        T = T * self.adjust_translation        
        R = rotation if rotation is not None else self.rotation.expand(b, 3, 3)
        Rt = torch.zeros(b, 4, 4).to(T)
        Rt[:, :3, :3] = R
        Rt[:, 3, :3] = T
        Rt[:, 3, 3] = 1.0
        inv_Rt = torch.inverse(Rt)
        
        fx, fy = self.focal_length
        px, py = self.principal_point
        if image is not None:            
            w, h = image.shape[-1], image.shape[-2]
            px, py = w / 2.0, h / 2.0
        elif nominal_image is not None:
            w, h = self.resolution
            ow, oh = nominal_image.shape[-1], nominal_image.shape[-2]
            fx, fy = fx * (w / ow), fy * (h / oh)
            if not self.has_nominal_principal:
                px, py = (px / ow) * w, (py / oh) * h
            else:
                px, py = 0.5 * w, 0.5 * h            
        else:
            w, h = self.resolution
        cameras = PerspectiveCameras(
            [(fx, fy)], [(px, py)],  
            R=inv_Rt[:, :3, :3],  T=inv_Rt[:, 3, :3], 
            in_ndc=False, image_size=[(h, w)]
        ).to(vertices.device)
        raster_settings = RasterizationSettings(
            image_size=(h, w), 
            blur_radius=np.log(1. / self.raster_blur_radius - 1.) * self.blend_params.sigma, 
            faces_per_pixel=self.faces_per_pixel, bin_size=self.raster_bin_size,
            perspective_correct=True, max_faces_per_bin=self.max_faces_per_bin
        )

        silhouette_renderer = MeshRenderer(
            rasterizer=MeshRasterizer(
                cameras=cameras, 
                raster_settings=raster_settings
            ),
            shader=SoftSilhouetteShader(blend_params=self.blend_params)
        )        
        textures = TexturesVertex(verts_features=torch.ones_like(vertices)) # [B, V, 3]
        mesh = Meshes(verts=vertices, faces=faces, textures=textures)
        color = silhouette_renderer(meshes_world=mesh)
        return color[:, np.newaxis, :, :, 3]


#NOTE: this sub-package code has been adapted from https://github.com/jonbarron/robust_loss_pytorch

from moai.supervision.losses.regression.robust.adaptive.learnable import (
    Barron,
    Student
) 

__all__ = [
    'Barron',
    'Student',
]


from moai.utils.arguments import assert_numeric

import torch
import logging

log = logging.getLogger(__name__)

__all__ = ["NormalPrior"]

class NormalPrior(torch.nn.Module):
    def __init__(self,
        mode:       str='standard', # ['standard', 'softplus']
        beta:       float=1.0,
        threshold:  float=20.0
    ) -> None:
        super(NormalPrior, self).__init__()
        assert_numeric(log, 'beta', beta, 0.0, None)
        assert_numeric(log, 'threshold', threshold, 0.0, None)
        self.softplus = torch.nn.Softplus(beta, threshold)
        self.prior= self.prior_with_softplus if mode == 'softplus'\
                        else self.prior_with_eps

    def prior_with_eps(self,
        mu:     torch.Tensor,
        logvar: torch.Tensor,
    ) -> torch.Tensor:
        std = torch.exp(0.5*logvar)
        eps = torch.rand_like(std)
        return mu + eps*std

    def prior_with_softplus(self,
        mu:     torch.Tensor,
        logvar: torch.Tensor,
    ) -> torch.Tensor:
        z = torch.distributions.normal.Normal(
            mu,
            self.softplus(logvar)
        )
        return z.rsample()

    def forward(self,
        mu:     torch.Tensor,
        logvar: torch.Tensor,
    ) -> torch.Tensor:
        return self.prior(mu, logvar)


from posixpath import lexists
import torch
import hydra.utils as hyu
import omegaconf.omegaconf
import typing
import toolz
import numpy as np
import logging

log = logging.getLogger(__name__)

__all__ = ["Interleaved"]

class Interleaved(torch.utils.data.Dataset):
    r"""Dataset creation by sampling from multiple datasets.

    This class is useful to assemble different existing datasets.

    Args:
        datasets (sequence): DictConfig of datasets to be concatenated
        probabilities (sequence): DictConfig of probabilities to create sampling
        size (int): Output size of concatenated dataset
    """
    def __init__(self,
        datasets:       omegaconf.DictConfig,
        size: int,
        probabilities:  omegaconf.DictConfig ,#typing.List[float]=[],
        augmentation:   omegaconf.DictConfig=None,
        extracted_keys: typing.List[str]=[],
    ):
        super().__init__()
        self.datasets = {}

        missing_datasets = set(datasets.keys()) - set(probabilities.keys())

        for key in missing_datasets:
            log.warning(f"Probability have not been assigned for dataset {key}."
                " and thus will be ingored for sampling.")
        
        if np.sum(list(probabilities.values())) != 1.0:
            log.warning(f"Probabilities do not sum up to unity ({np.sum(list(probabilities.values()))}), they will be normalized to unity.")
            probabilities = toolz.valmap(lambda x: x /np.sum(list(probabilities.values())),probabilities)
        if augmentation is not None:
            for (key, dataset) in zip(datasets.keys(),datasets.values()):
                self.datasets[key] = (hyu.instantiate(
                    augmentation,
                    hyu.instantiate(dataset)
                ))
        else:
            from moai.data.augmentation import NoOp
            for (key,dataset) in zip(datasets.keys(),datasets.values()):
                self.datasets[key] = (NoOp(hyu.instantiate(dataset)))
        self.lengths = toolz.valmap(lambda x: x.__len__(),self.datasets)
        self.keys = [k.split('.') for k in extracted_keys]
        self.probabilities = probabilities
        self.size = size
        for d,p in zip(probabilities.keys(),probabilities.values()):
            log.info(f"Drawing samples from {d} with probability {'{:.2f}'.format(p * 100)}%")

    
    def __len__(self) -> int:
        return self.size

    def __getitem__(self, index: int) -> typing.Dict[str, torch.Tensor]:
        dataset_idx =  np.random.choice(list(self.probabilities.keys()), p = list(self.probabilities.values())) #sample from dataset with probabilities
        index = np.random.randint(0,self.lengths[dataset_idx]) 
        item = self.datasets[dataset_idx][index] 
        if self.keys:            
            out = {}
            for k in self.keys:
                out = toolz.assoc_in(out, k, toolz.get_in(k, item))
            return out
        else:
            return item


import itertools
import moai.networks.lightning as minet
import moai.nn.convolution as mic
import moai.nn.sampling.spatial.downsample as mids
import moai.nn.sampling.spatial.upsample as mius

import torch
import hydra.utils as hyu
import omegaconf.omegaconf as omegaconf
import typing
import logging
import toolz

log = logging.getLogger(__name__)

__all__ = ["UNet"]

class UNet(minet.FeedForward):
    class DefaultSkip(torch.nn.Module):
        def __init__(self):
            super(UNet.DefaultSkip, self).__init__()

        def forward(self, 
            enc:    torch.Tensor,
            dec:    torch.Tensor,
            gate:   torch.Tensor=None,
        ) -> torch.Tensor:
            return torch.cat([dec, enc], dim=1)

    def __init__(self,
        configuration:  omegaconf.DictConfig,
        modules:        omegaconf.DictConfig=None,
        data:           omegaconf.DictConfig=None,
        parameters:     omegaconf.DictConfig=None,
        feedforward:    omegaconf.DictConfig=None,
        monads:         omegaconf.DictConfig=None,
        supervision:    omegaconf.DictConfig=None,
        validation:     omegaconf.DictConfig=None,
        visualization:  omegaconf.DictConfig=None,
        export:         omegaconf.DictConfig=None,
    ):
        super(UNet, self).__init__(
            data=data, parameters=parameters,
            feedforward=feedforward, monads=monads,
            supervision=supervision, validation=validation,
            export=export, visualization=visualization,            
        )        
        # Encoder
        convolution = configuration.encoder.convolution
        activation = configuration.encoder.activation
        downscale = configuration.encoder.downscale
        block_features = configuration.block_features
        down_block_convs = configuration.encoder.blocks
        in_features = configuration.in_features
        self.enc = torch.nn.ModuleDict()
        self.down = torch.nn.ModuleDict()
        for f, b in zip(block_features, down_block_convs):
            enc_block = torch.nn.Sequential()
            for i in range(b): #TODO: can modify encoder conv block
                enc_block.add_module(f'enc{f}_{i}', mic.make_conv_block(
                    block_type='conv2d',
                    convolution_type=convolution.type,
                    in_features=in_features if i == 0 else f, 
                    out_features=f,
                    activation_type=activation.type,
                    convolution_params=convolution.params,
                    activation_params=activation.params
                ))
            in_features = f
            self.enc.add_module(f'enc_block{f}', enc_block)            
            self.down.add_module(f'down{f}', mids.make_downsample(
                downscale_type=downscale.type,
                features=f,
                kernel_size=3 if downscale.type == 'maxpool2d_aa' else 2,
                **(downscale.params if downscale.params is not None else {})
            ))        
        # Bottleneck
        bottleneck = configuration.bottleneck
        self.bottleneck = torch.nn.Sequential()
        for i in range(bottleneck.blocks): #TODO: can modify bottleneck block
            self.bottleneck.add_module(f'bottleneck{i}', mic.make_conv_block(
                block_type='conv2d',
                convolution_type=bottleneck.convolution.type,
                in_features=block_features[-1] if i == 0 else bottleneck.features, 
                out_features=bottleneck.features,
                activation_type=bottleneck.activation.type,
                convolution_params=bottleneck.convolution.params,
                activation_params=bottleneck.activation.params
            ))
        gate = toolz.get_in(['gate'], modules)        
        self.gate = hyu.instantiate(gate, bottleneck.features) if gate else torch.nn.Identity()        
        # Decoder
        convolution = configuration.decoder.convolution
        activation = configuration.decoder.activation
        upscale = configuration.decoder.upscale
        skip = toolz.get_in(['skip'], modules)
        expansion = toolz.get_in(['skip', 'expansion'], modules) or 2.0        
        up_block_convs = configuration.decoder.blocks
        up_block_convs = up_block_convs or reversed(down_block_convs)
        self.dec = torch.nn.ModuleDict()
        self.up = torch.nn.ModuleDict()
        self.skip = torch.nn.ModuleDict()
        skip_params = configuration.skip or {}
        if not isinstance(skip_params, typing.Sized) or len(skip_params) <= 1:
            skip_params = list(itertools.repeat(skip_params, len(down_block_convs)))
        for f, b, sp in zip(reversed(block_features), up_block_convs, skip_params):  #TODO: can modify upscale block
            self.up.add_module(f'up{f}', torch.nn.Sequential(
                mius.make_upsample(
                    upscale_type=upscale.type,
                    features=f * 2,
                    **(upscale.params if upscale.params is not None else {}),
                ),
                mic.make_conv_1x1(
                    convolution_type=convolution.type,
                    in_channels=f * 2, out_channels=f,
                ) if upscale.type == 'upsample2d' else torch.nn.Identity()
            ))#TODO: need to downscale features too
            #NOTE: one case is deconv2d and the other is upscale and project
            #NOTE: we now only use upscale and project
            self.skip.add_module(f'skip{f}', 
                hyu.instantiate(skip, f, **sp) if skip is not None else UNet.DefaultSkip()
            )
            dec_block = torch.nn.Sequential()
            skip_features = int(f * expansion)
            for i in range(b):
                dec_block.add_module(f'dec{f}_{i}', mic.make_conv_block(
                    block_type='conv2d',
                    convolution_type=convolution.type,
                    in_features=skip_features if i == 0 else f, 
                    out_features=f,
                    activation_type=activation.type,
                    convolution_params=convolution.params,
                    activation_params=activation.params
                ))
            self.dec.add_module(f'dec_block{f}', dec_block)        
        # Prediction
        prediction = configuration.prediction
        out_features = configuration.out_features
        self.pred = mic.make_conv_block('conv2d',
            convolution_type=prediction.convolution.type,
            in_features=block_features[0], out_features=out_features,
            activation_type=prediction.activation.type,
            convolution_params=prediction.convolution.params,
            activation_params=prediction.activation.params,
        )
        self.input = configuration.input
        self.output = configuration.output
        
    def forward(self, 
        td: typing.Dict[str, torch.Tensor]
    ) -> typing.Dict[str, torch.Tensor]:
        for i, o in zip(self.input, self.output):
            data = td[i]
            features = [data]
            skipped = []
            for e, d in zip(self.enc.values(), self.down.values()):
                skipped.append(e(features[-1]))
                features.append(d(skipped[-1]))
            bottleneck = self.bottleneck(features[-1])
            skipped.reverse()
            features = [bottleneck]
            gate = self.gate(bottleneck)
            for i, (u, s, d) in enumerate(zip(
                self.up.values(),
                self.skip.values(),
                self.dec.values()
            )):
                up = u(features[-1])
                skip = s(skipped[i], up, gate)
                features.append(d(skip))
            out = self.pred(features[-1])
            td[o] = out
        return td


from moai.utils.arguments import assert_numeric
from moai.monads.execution.cascade import _create_accessor
from collections.abc import Callable

import scenepic
import httpwatcher
import typing
import torch
import logging
import colour
import toolz
import math
import os
import numpy as np
import cv2

log = logging.getLogger(__name__)

__all__ = ['Mesh']


class Mesh(Callable):    
    
    __LAMBDA_MAP__ = { '': lambda _: None, 'skeleton': lambda _: 'skeleton'}

    def __init__(self,
        vertices:         typing.Union[str, typing.Sequence[str]],
        faces:            typing.Union[str, typing.Sequence[str]],
        canvas:           typing.Union[int, typing.Sequence[int]],
        layer:            typing.Union[int, typing.Sequence[int]],
        color:            typing.Union[str, typing.Sequence[str]],
        batch_percentage: float=1.0,
        width:            int=600,
        height:           int=400,
        point_size:       float=0.1,
        name:             str="default",
        skeleton:         typing.Sequence[int]=None,  
        joint_radius:      float=0.05,
        start_thickness:   float=0.01,
        end_thickness:     float=0.05,
        add_wireframe:     bool=False,
    ):
        self.name, self.point_size = name, point_size
        self.vertices = [vertices] if isinstance(vertices, str) else list(vertices)
        self.faces = [faces] if isinstance(faces, str) else list(faces)
        self.vertex_accessors = [_create_accessor(k) for k in self.vertices]        
        self.face_accesors = [(_create_accessor(k) if k not in Mesh.__LAMBDA_MAP__ else Mesh.__LAMBDA_MAP__[k]) for k in self.faces]
        self.ids = [canvas] if isinstance(canvas, int) else list(canvas)
        self.layers = [layer] if isinstance(layer, int) else list(layer)
        self.colors = [colour.Color(color)] if isinstance(color, str) else list(colour.Color(c) for c in color)
        self.batch_percentage = batch_percentage
        assert_numeric(log, 'batch percentage', batch_percentage, 0.0, 1.0)        
        self.WH = (width, height)
        scenepic_folder = os.path.join(os.getcwd(), 'scenepic')
        self.joint_radius = joint_radius
        self.kintree = skeleton
        self.start_thickness = start_thickness
        self.end_thickness = end_thickness
        self.add_wireframe = add_wireframe
        os.makedirs(scenepic_folder, exist_ok=True)
        log.info(f"Scenepic visualization enabled @ {scenepic_folder}.")
        log.warning(f"[scenepic]: For automatic refreshing @ `http://localhost:5555' use `httpwatcher -r {scenepic_folder}` (`pip install httpwatcher` if not available)")

    def _draw_skeleton(self, mesh, kpts):
        for nk, (i, j) in enumerate(self.kintree):
            mesh.add_thickline(
                start_point = kpts[i], 
                end_point = kpts[j],
                start_thickness = self.start_thickness,
                end_thickness  = self.end_thickness,
                add_wireframe = self.add_wireframe
                )
        for nj in range(kpts.shape[0]):
            tranform_sphere = np.identity(4)
            tranform_sphere[:3,3] = kpts[nj]
            tranform_sphere[
                (np.diag_indices_from(tranform_sphere)[0][:3]),
                (np.diag_indices_from(tranform_sphere)[1][:3])
                ] = self.joint_radius
            mesh.add_sphere(transform = tranform_sphere,add_wireframe = self.add_wireframe)
    
    
    def __call__(self, 
        tensors:    typing.Dict[str, torch.Tensor],
        step:       typing.Optional[int]=None,
    ) -> None:
        meshes = {}
        scene = scenepic.Scene()
        canvases = [
            scene.create_canvas_3d(width=self.WH[0], height=self.WH[1]) 
            for _ in toolz.unique(self.ids)
        ]
        
        for n, v, f, c, l, id in zip(self.vertices, self.vertex_accessors,
            self.face_accesors, self.colors, self.layers, self.ids
        ):
            draw_skeleton = False
            vertices = v(tensors).detach().cpu().numpy()
            faces = f(tensors)
            if faces is not None and faces != 'skeleton':
                faces = faces.detach().cpu().numpy()
            else:
                if faces == 'skeleton':
                    draw_skeleton = True
            b = math.ceil(self.batch_percentage * vertices.shape[0])
            if id not in meshes:
                meshes[id] = []
            for i in range(b):
                mesh = scene.create_mesh(mesh_id=f"{n}_{i}",
                    shared_color=scenepic.Color(*c.rgb), layer_id=f"{l}",
                )
                if draw_skeleton:
                    self._draw_skeleton(mesh,vertices[i])
                else:
                    if faces is not None:
                        mesh.add_mesh_without_normals(vertices[i], faces[i])
                    else:
                        mesh.add_sphere()
                        mesh.apply_transform(scenepic.Transforms.Scale(self.point_size)) 
                        mesh.enable_instancing(positions=vertices[i]) 
                meshes[id].append(mesh)
        for k, m in meshes.items():
            frame = canvases[k].create_frame()
            grouped = toolz.groupby(lambda x: x.layer_id, m)
            for l, m in grouped.items():
                for j, mesh in enumerate(m):
                    xform = scenepic.Transforms.Translate(j * np.array([1.0, 0.0, 0.0]))
                    frame.add_mesh(mesh, xform)
            frame.set_layer_settings(dict(
                (str(n), {'opacity': 0.5}) for n in toolz.unique(self.layers)
            ))
        scene.link_canvas_events(*canvases)
        scene.save_as_html(os.path.join('scenepic', "index.html"), title=f"{self.name}")


from moai.supervision.losses.regression.L1 import L1
from moai.supervision.losses.regression.L2 import L2
from moai.supervision.losses.regression.cosine_distance import CosineDistance
from moai.supervision.losses.regression.distance import Distance
from functools import partial

LogL1 = partial(L1, mode='log')
LnL1 = partial(L1, mode='ln')

__all__ = [
    "CosineDistance",
    "L1",
    "LogL1",
    "LnL1",
    "L2",
    "Distance",
]


from moai.data.datasets.human.genebody import GeneBody
from moai.data.datasets.human.amass import AMASS
from moai.data.datasets.human.fit3d import Fit3D
from moai.data.datasets.human.thuman2 import THuman2

__all__ = [
    'GeneBody',
    'AMASS',
    'Fit3D',
    'THuman2',
]


from kornia.filters import (
    MedianBlur
)

import torch
import logging
import functools
import typing

log = logging.getLogger(__name__)

__all__ = ["Median2d"]


class Median2d(MedianBlur):
    def __init__(self,
        kernel_size:                typing.Tuple[int, int]=(5, 5),
        # border_type:                str='reflect'
    ):
        super(Median2d, self).__init__(
            kernel_size=tuple(kernel_size),
            # border_type=border_type
        )

    def forward(self, image: torch.Tensor) -> torch.Tensor:
        if len(image.shape) < 4:
            image = image.unsqueeze(0).clone()
        return super(Median2d, self).forward(image)